import numpy as np
from pyscf import lib
from pyscf.df.grad import uks as uks_grad_df
from pyscf.dft import gen_grid, numint
from pyscf.grad import rks as rks_grad
from pyscf.grad import uks as uks_grad
from pyscf.grad.rks import grids_noresponse_cc, grids_response_cc
from pyscf.lib import logger


def get_veff(ks_grad, mol=None, dm=None):
    """
    First order derivative of DFT effective potential matrix (wrt electron coordinates)

    Args:
        ks_grad : grad.uhf.Gradients or grad.uks.Gradients object
    """
    if mol is None:
        mol = ks_grad.mol
    if dm is None:
        dm = ks_grad.base.make_rdm1()
    t0 = (logger.process_clock(), logger.perf_counter())

    mf = ks_grad.base
    ni = mf._numint
    grids, nlcgrids = rks_grad._initialize_grids(ks_grad)

    ni = mf._numint
    mem_now = lib.current_memory()[0]
    max_memory = max(2000, ks_grad.max_memory * 0.9 - mem_now)
    if ks_grad.grid_response:
        exc, vxc = get_vxc_full_response(
            ni, mol, grids, mf.xc, dm, max_memory=max_memory, verbose=ks_grad.verbose
        )
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = rks_grad.get_nlc_vxc_full_response(
                ni,
                mol,
                nlcgrids,
                xc,
                dm[0] + dm[1],
                max_memory=max_memory,
                verbose=ks_grad.verbose,
            )
            exc += enlc
            vxc += vnlc
        logger.debug1(ks_grad, "sum(grids response) %s", exc.sum(axis=0))
    else:
        exc, vxc = get_vxc(
            ni, mol, grids, mf.xc, dm, max_memory=max_memory, verbose=ks_grad.verbose
        )
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = rks_grad.get_nlc_vxc(
                ni,
                mol,
                nlcgrids,
                xc,
                dm[0] + dm[1],
                max_memory=max_memory,
                verbose=ks_grad.verbose,
            )
            vxc += vnlc
    t0 = logger.timer(ks_grad, "vxc", *t0)

    has_df = hasattr(mf, "with_df") and mf.with_df is not None
    if not has_df:
        # no density fitting
        if not ni.libxc.is_hybrid_xc(mf.xc):
            vj = ks_grad.get_j(mol, dm)
            vxc += vj[0] + vj[1]
        else:
            raise NotImplementedError("Hybrid DFT with CIDER functionals")

        return lib.tag_array(vxc, exc1_grid=exc)
    else:
        # density fitting is used, TODO check this always works more thoroughly
        if not ni.libxc.is_hybrid_xc(mf.xc):
            vj = ks_grad.get_j(mol, dm)
            vxc += vj[0] + vj[1]
            if ks_grad.auxbasis_response:
                e1_aux = vj.aux.sum((0, 1))
        else:
            raise NotImplementedError("Hybrid DFT with CIDER functionals")

        if ks_grad.auxbasis_response:
            logger.debug1(ks_grad, "sum(auxbasis response) %s", e1_aux.sum(axis=0))
            vxc = lib.tag_array(vxc, exc1_grid=exc, aux=e1_aux)
        else:
            vxc = lib.tag_array(vxc, exc1_grid=exc)
        return vxc


def get_vxc(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    if not ni.settings.nldf_settings.is_empty:
        return get_vxc_nldf(
            ni,
            mol,
            grids,
            xc_code,
            dms,
            relativity=relativity,
            hermi=hermi,
            max_memory=max_memory,
            verbose=verbose,
        )
    ni.timer.start("nr_uks")
    if relativity != 0:
        raise NotImplementedError

    if not ni.settings.sdmx_settings.is_empty:
        raise NotImplementedError("SDMX forces")
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError("NLOF forces")

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 2)

    dma, dmb = numint._format_uks_dm(dms)
    assert dma.ndim == 2
    nao = dma.shape[-1]
    make_rhoa, nset = ni._gen_rho_evaluator(mol, dma, hermi, False, grids)[:2]
    make_rhob = ni._gen_rho_evaluator(mol, dmb, hermi, False, grids)[0]
    ao_loc = mol.ao_loc_nr()
    extra_ao = 0
    ao_deriv = 2

    vmat = np.zeros((2, 3, nao, nao))
    for ao, mask, weight, coords in ni.block_loop(
        mol, grids, nao, ao_deriv, max_memory=max_memory, extra_ao=extra_ao
    ):
        rho_a = make_rhoa(0, ao, mask, xctype)
        rho_b = make_rhob(0, ao, mask, xctype)
        rho = (rho_a, rho_b)
        ni.timer.start("xc cider")
        exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
            xc_code, rho, None, None, deriv=1, xctype=xctype
        )[:2]
        ni.timer.stop("xc cider")
        assert vxc_nldf is None
        wv = weight * vxc
        wv[:, 0] *= 0.5
        rks_grad._gga_grad_sum_(vmat[0], mol, ao, wv[0], mask, ao_loc)
        rks_grad._gga_grad_sum_(vmat[1], mol, ao, wv[1], mask, ao_loc)
        if xctype == "MGGA":
            wv[:, 4] *= 0.5
            rks_grad._tau_grad_dot_(vmat[0], mol, ao, wv[0, 4], mask, ao_loc, True)
            rks_grad._tau_grad_dot_(vmat[1], mol, ao, wv[1, 4], mask, ao_loc, True)

    exc = np.zeros((mol.natm, 3))
    # - sign because nabla_X = -nabla_x
    return exc, -vmat


def get_vxc_nldf(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    ni.timer.start("nr_uks")
    if relativity != 0:
        raise NotImplementedError

    if not ni.settings.sdmx_settings.is_empty:
        raise NotImplementedError("SDMX forces")
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError("NLOF forces")

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 2)

    dma, dmb = numint._format_uks_dm(dms)
    assert dma.ndim == 2
    nao = dma.shape[-1]
    make_rhoa, nset = ni._gen_rho_evaluator(mol, dma, hermi, False, grids)[:2]
    make_rhob = ni._gen_rho_evaluator(mol, dmb, hermi, False, grids)[0]
    assert nset == 1
    ao_loc = mol.ao_loc_nr()
    nrho = 5 if xctype == "MGGA" else 4
    if not ni.settings.nlof_settings.is_empty:
        nrho += ni.settings.nlof_settings.nrho
    rhoa_full = np.zeros((nrho, grids.weights.size), dtype=np.float64, order="C")
    rhob_full = np.zeros((nrho, grids.weights.size), dtype=np.float64, order="C")
    wva_full = np.zeros((nrho, grids.weights.size), dtype=np.float64, order="C")
    wvb_full = np.zeros((nrho, grids.weights.size), dtype=np.float64, order="C")

    num_nldf = ni.settings.nldf_settings.nfeat
    vxc_nldf_full = np.zeros(
        (2, num_nldf, grids.weights.size), dtype=np.float64, order="C"
    )

    def block_loop(ao_deriv):
        ip0 = 0
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory
        ):
            ip1 = ip0 + weight.size
            yield ip0, ip1, ao, mask, weight, coords
            ip0 = ip1

    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    ao_deriv = 1
    for ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        rhoa_full[:, ip0:ip1] = make_rhoa(0, ao, mask, xctype)
        rhob_full[:, ip0:ip1] = make_rhob(0, ao, mask, xctype)

    par_atom = False
    extra_ao = ni.nldfgen.get_extra_ao()
    if par_atom:
        raise NotImplementedError
    else:
        nldf_feat = np.stack(
            [
                ni.nldfgen.get_features(rhoa_full, spin=0),
                ni.nldfgen.get_features(rhob_full, spin=1),
            ]
        )
        ip0 = 0
        for mask, weight, coords in ni.extra_block_loop(
            mol, grids, max_memory=max_memory, extra_ao=extra_ao
        ):
            ip1 = ip0 + weight.size
            rho_a = rhoa_full[:, ip0:ip1]
            rho_b = rhob_full[:, ip0:ip1]
            rho = (rho_a, rho_b)
            exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                xc_code,
                rho,
                nldf_feat[..., ip0:ip1],
                None,
                deriv=1,
                xctype=xctype,
            )[:2]
            vxc_nldf_full[..., ip0:ip1] = vxc_nldf * weight
            wva_full[:, ip0:ip1] = weight * vxc[0]
            wvb_full[:, ip0:ip1] = weight * vxc[1]
            ip0 = ip1
        wva_full[:, :] += ni.nldfgen.get_potential(vxc_nldf_full[0], spin=0)
        wvb_full[:, :] += ni.nldfgen.get_potential(vxc_nldf_full[1], spin=1)

    ao_deriv = 2
    vmat = np.zeros((2, 3, nao, nao))
    for ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        rho_a = np.ascontiguousarray(rhoa_full[..., ip0:ip1])
        rho_b = np.ascontiguousarray(rhob_full[..., ip0:ip1])
        rho = (rho_a, rho_b)
        wva = np.ascontiguousarray(wva_full[..., ip0:ip1])
        wvb = np.ascontiguousarray(wvb_full[..., ip0:ip1])
        wva[0] *= 0.5
        wvb[0] *= 0.5
        rks_grad._gga_grad_sum_(vmat[0], mol, ao, wva, mask, ao_loc)
        rks_grad._gga_grad_sum_(vmat[1], mol, ao, wvb, mask, ao_loc)
        if xctype == "MGGA":
            wva[4] *= 0.5
            wvb[4] *= 0.5
            rks_grad._tau_grad_dot_(vmat[0], mol, ao, wva[4], mask, ao_loc, True)
            rks_grad._tau_grad_dot_(vmat[1], mol, ao, wvb[4], mask, ao_loc, True)

    exc = np.zeros((mol.natm, 3))
    # - sign because nabla_X = -nabla_x
    return exc, -vmat


def get_vxc_full_response(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    if not ni.settings.nldf_settings.is_empty:
        return get_vxc_nldf_full_response(
            ni,
            mol,
            grids,
            xc_code,
            dms,
            relativity=relativity,
            hermi=hermi,
            max_memory=max_memory,
            verbose=verbose,
        )
    ni.timer.start("nr_uks")
    if relativity != 0:
        raise NotImplementedError

    if not ni.settings.sdmx_settings.is_empty:
        raise NotImplementedError("SDMX forces")
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError("NLOF forces")

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 2)

    dma, dmb = numint._format_uks_dm(dms)
    assert dma.ndim == 2
    nao = dma.shape[-1]
    make_rhoa, nset = ni._gen_rho_evaluator(mol, dma, hermi, False, grids)[:2]
    make_rhob = ni._gen_rho_evaluator(mol, dmb, hermi, False, grids)[0]
    ao_loc = mol.ao_loc_nr()
    ao_deriv = 2
    excsum = 0

    vmat = np.zeros((2, 3, nao, nao))
    for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask, cutoff=grids.cutoff)
        rho_a = make_rhoa(0, ao, mask, xctype)
        rho_b = make_rhob(0, ao, mask, xctype)
        rho = (rho_a, rho_b)
        ni.timer.start("xc cider")
        exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
            xc_code, rho, None, None, deriv=1, xctype=xctype
        )[:2]
        ni.timer.stop("xc cider")
        assert vxc_nldf is None
        wv = weight * vxc
        wv[:, 0] *= 0.5
        if xctype == "MGGA":
            wv[:, 4] *= 0.5

        vtmp = np.zeros((3, nao, nao))
        rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[0], mask, ao_loc)
        if xctype == "MGGA":
            rks_grad._tau_grad_dot_(vtmp, mol, ao, wv[0, 4], mask, ao_loc, True)
        vmat[0] += vtmp
        excsum += np.einsum("r,r,nxr->nx", exc, rho_a[0] + rho_b[0], weight1)
        excsum[atm_id] += np.einsum("xij,ji->x", vtmp, dms[0]) * 2

        vtmp = np.zeros((3, nao, nao))
        rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[1], mask, ao_loc)
        if xctype == "MGGA":
            rks_grad._tau_grad_dot_(vtmp, mol, ao, wv[1, 4], mask, ao_loc, True)
        vmat[1] += vtmp
        excsum[atm_id] += np.einsum("xij,ji->x", vtmp, dms[1]) * 2

    # - sign because nabla_X = -nabla_x
    return excsum, -vmat


def get_vxc_nldf_full_response(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    ni.timer.start("nr_uks")
    if relativity != 0:
        raise NotImplementedError

    if not ni.settings.sdmx_settings.is_empty:
        raise NotImplementedError("SDMX forces")
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError("NLOF forces")

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 2)

    dma, dmb = numint._format_uks_dm(dms)
    assert dma.ndim == 2
    nao = dma.shape[-1]
    make_rhoa, nset = ni._gen_rho_evaluator(mol, dma, hermi, False, grids)[:2]
    make_rhob = ni._gen_rho_evaluator(mol, dmb, hermi, False, grids)[0]
    ao_loc = mol.ao_loc_nr()
    ao_deriv = 2
    excsum = 0

    vmat = np.zeros((2, 3, nao, nao))
    nrho = 5 if xctype == "MGGA" else 4
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError

    num_nldf = ni.settings.nldf_settings.nfeat
    vxc_nldf_full = np.zeros(
        (2, num_nldf, grids.weights.size), dtype=np.float64, order="C"
    )

    ga_loc = ni.nldfgen.interpolator.grid_loc_atom
    assert ga_loc is not None
    rho_full = np.zeros((2, nrho, ga_loc[-1]), dtype=np.float64, order="C")
    wv_full = np.zeros((2, nrho, ga_loc[-1]), dtype=np.float64, order="C")
    all_coords = np.zeros((ga_loc[-1], 3), dtype=np.float64, order="C")
    exc = np.zeros(ga_loc[-1], dtype=np.float64, order="C")

    ao_deriv = 1
    for atm_id, (coords, weight) in enumerate(grids_noresponse_cc(grids)):
        ip0, ip1 = ga_loc[atm_id : atm_id + 2]
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask, cutoff=grids.cutoff)
        rho_full[0, :, ip0:ip1] = make_rhoa(0, ao, mask, xctype)
        rho_full[1, :, ip0:ip1] = make_rhob(0, ao, mask, xctype)
        all_coords[ip0:ip1] = coords

    par_atom = False
    extra_ao = ni.nldfgen.get_extra_ao()
    if ni.has_sdmx:
        raise NotImplementedError
    if par_atom:
        raise NotImplementedError
    else:
        nldf_feat = np.stack(
            [
                ni.nldfgen.get_features(
                    rho_full[0], map_grids=False, grad_mode=True, spin=0
                ),
                ni.nldfgen.get_features(
                    rho_full[1], map_grids=False, grad_mode=True, spin=1
                ),
            ]
        )
        ip0 = 0
        for mask, weight, coords in ni.extra_block_loop(
            mol, grids, max_memory=max_memory, extra_ao=extra_ao
        ):
            ip1 = ip0 + weight.size
            ip1 = min(ip1, grids.grids_indexer.idx_map.size)
            weight = weight[: ip1 - ip0]
            inds = grids.grids_indexer.idx_map[ip0:ip1]
            rho = (rho_full[0][..., inds], rho_full[1][..., inds])
            exc[inds], (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                xc_code,
                rho,
                np.ascontiguousarray(nldf_feat[..., ip0:ip1]),
                None,
                deriv=1,
                xctype=xctype,
            )[:2]
            vxc_nldf_full[..., ip0:ip1] = vxc_nldf * weight
            wv_full[..., inds] = weight * vxc
            ip0 = ip1
        wvtmp, cidergg_g, excsum = ni.nldfgen.get_potential(
            vxc_nldf_full[0], map_grids=False, grad_mode=True, spin=0
        )
        wv_full[0] += wvtmp
        wvtmp, _cidergg_g, _excsum = ni.nldfgen.get_potential(
            vxc_nldf_full[1], map_grids=False, grad_mode=True, spin=1
        )
        cidergg_g += _cidergg_g
        excsum += _excsum
        wv_full[1] += wvtmp

    ao_deriv = 2
    for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask, cutoff=grids.cutoff)
        ip0, ip1 = ga_loc[atm_id : atm_id + 2]

        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask, cutoff=grids.cutoff)
        rho_a = np.ascontiguousarray(rho_full[0, :, ip0:ip1])
        rho_b = np.ascontiguousarray(rho_full[1, :, ip0:ip1])
        wv = np.ascontiguousarray(wv_full[..., ip0:ip1])
        wv[:, 0] *= 0.5
        if xctype == "MGGA":
            wv[:, 4] *= 0.5

        vtmp = np.zeros((3, nao, nao))
        rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[0], mask, ao_loc)
        if xctype == "MGGA":
            rks_grad._tau_grad_dot_(vtmp, mol, ao, wv[0, 4], mask, ao_loc, True)
        vmat[0] += vtmp
        excsum += np.einsum("r,r,nxr->nx", exc[ip0:ip1], rho_a[0] + rho_b[0], weight1)
        excsum[atm_id] += np.einsum("xij,ji->x", vtmp, dms[0]) * 2

        vtmp = np.zeros((3, nao, nao))
        rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[1], mask, ao_loc)
        if xctype == "MGGA":
            rks_grad._tau_grad_dot_(vtmp, mol, ao, wv[1, 4], mask, ao_loc, True)
        vmat[1] += vtmp
        excsum[atm_id] += np.einsum("xij,ji->x", vtmp, dms[1]) * 2

        excsum += np.dot(weight1, cidergg_g[ip0:ip1])
        ip0 = ip1
    # - sign because nabla_X = -nabla_x
    return excsum, -vmat


class Gradients(uks_grad.Gradients):

    get_veff = get_veff


class DFGradients(uks_grad_df.Gradients):

    get_veff = get_veff



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

"""
This module is for fast evaluation of features and their derivatives with
respect to orbital occupations and rotations. Only for b and d features
"""

import numpy as np
from pyscf import lib
from pyscf.dft.numint import NumInt

from ciderpress.dft.plans import FracLaplPlan, SemilocalPlan
from ciderpress.dft.settings import (
    FracLaplSettings,
    NLDFSettings,
    SDMXBaseSettings,
    SemilocalSettings,
)
from ciderpress.pyscf.analyzers import UHFAnalyzer
from ciderpress.pyscf.frac_lapl import FLNumInt
from ciderpress.pyscf.gen_cider_grid import CiderGrids
from ciderpress.pyscf.nldf_convolutions import DEFAULT_CIDER_LMAX, PyscfNLDFGenerator
from ciderpress.pyscf.sdmx_slow import EXXSphGenerator

NLDF_VERSION_LIST = ["i", "j", "ij", "k"]


def get_descriptors(analyzer, settings, orbs=None, **kwargs):
    """
    Get the descriptors for version b or d (l=0 only). Analyzer must
    contain mo_occ, mo_energy, and mo_coeff attributes.
    Can get feature derivatives with respect to orbital occupation using orbs.
    Spin polarization is determined automatically from the analyzer class type.

    orbs format:
        orbs is a dictionary of lists or (for a spin-polarized system) a
        2-tuple of dictionary of lists. Each key is either 'B', 'O', or 'U',
        with the value being a list of orbital indexes with respect to which to
        compute occupation derivatives of the features. 'B' indexes up from the
        bottom, i.e. the lowest-energy core states. 'O' indexes down from the
        highest occupied molecular orbital. 'U' indexes up from the lowest
        unoccupied molecular orbital. 0-indexing is used in all cases.
        For the spin-polarized case, if orbs is a single dictionary rather than
        a 2-tuple, orbital indexing is performed over the combined set of
        orbitals.

    Args:
        analyzer (RHFAnalyzer or UHFAnalyzer): analyzer object containing the
            desired molecule and density matrix/orbitals for which to compute
            features and their derivatives.
        version: 'b', 'd', or 'l', feature version to evaluate. 'l' stands
            for local and computes [rho, drho_x, drho_y, drho_z, tau]
        orbs: Orbital index dictionary for computing orbital derivatives.
        **kwargs: Any settings one wants to change from the default for the
                  PyscfNLDFGenerator object.

    Returns:
        desc (np.ndarray (nspin, nfeat, ngrid)): Descriptors for the system
        ddesc (dict(str : dict(int : array(nfeat, ngrid)))) or (dict, dict):
            Derivatives of the features with respect to the orbitals given by
            orbs, stored in a nested dictionary first by indexing style
            (B, O, or U) and then by an orbital index. If orbs was a 2-tuple,
            ddesc is a 2-tuple. Index format is the same as orbs, so a
            U-indexed orbital in orbs is U-indexed in ddesc. If calculation
            was spin-polarized but orbs was not spin-separated, each value in
            the ddesc dictionary is a 2-tuple (spin, deriv_array)
        eigvals (dict(str : dict(int : float)) or (dict, dict):
            Eigenvalues of the relevant orbitals. These are important as
            baseline energy derivatives, assuming the baseline functional
            is the same as the reference functional for the eigenvalues.
    """
    spinpol = isinstance(analyzer, UHFAnalyzer)
    mo_occ = analyzer.mo_occ
    mo_energy = analyzer.mo_energy
    mo_coeff = analyzer.mo_coeff
    if orbs is None:
        labels, en_list, sep_spins = None, None, None
        if spinpol:
            coeffs = [None, None]
        else:
            coeffs = None
    else:
        labels, coeffs, en_list, sep_spins = get_labels_and_coeffs(
            orbs, mo_coeff, mo_occ, mo_energy
        )

    if isinstance(settings, str):
        if settings != "l":
            raise ValueError("Settings must be Settings object or letter l")
        my_desc_getter = _density_getter
    elif isinstance(settings, SemilocalSettings):
        my_desc_getter = _sl_desc_getter
    elif isinstance(settings, NLDFSettings):
        my_desc_getter = _nldf_desc_getter
    elif isinstance(settings, FracLaplSettings):
        my_desc_getter = _fl_desc_getter
    elif isinstance(settings, SDMXBaseSettings):
        my_desc_getter = _sdmx_desc_getter
    else:
        raise ValueError("Unsupported settings")

    if spinpol:
        desc = []
        for s in range(2):
            desc.append(
                my_desc_getter(
                    analyzer.mol,
                    analyzer.grids,
                    2 * analyzer.rdm1[s],
                    settings,
                    coeffs=coeffs[s],
                    **kwargs
                )
            )
        desc_a, desc_b = desc
        if orbs is not None:
            desc_a, deriv_arr_a = desc_a
            desc_b, deriv_arr_b = desc_b
            deriv_arr_a *= 2.0
            deriv_arr_b *= 2.0
            if sep_spins:
                ddesc_a, eigvals_a = unpack_feature_derivs(
                    orbs[0], labels[0], deriv_arr_a, en_list[0]
                )
                ddesc_b, eigvals_b = unpack_feature_derivs(
                    orbs[1], labels[1], deriv_arr_b, en_list[1]
                )
                return (
                    np.stack([desc_a, desc_b]),
                    (ddesc_a, ddesc_b),
                    (eigvals_a, eigvals_b),
                )
            else:
                deriv_arr = [deriv_arr_a, deriv_arr_b]
                ddesc = {}
                eigvals = {}
                for k in list(orbs.keys()):
                    ddesc[k] = {}
                    eigvals[k] = {}
                for s in range(2):
                    ddesc_tmp, eigvals_tmp = unpack_feature_derivs(
                        orbs, labels[s], deriv_arr[s], en_list[s]
                    )
                    for k in list(orbs.keys()):
                        ddesc[k].update(
                            {kk: (s, vv) for kk, vv in ddesc_tmp[k].items()}
                        )
                        eigvals[k].update(
                            {kk: (s, vv) for kk, vv in eigvals_tmp[k].items()}
                        )
                return np.stack([desc_a, desc_b]), ddesc, eigvals
        else:
            return np.stack([desc_a, desc_b])
    else:
        desc = my_desc_getter(
            analyzer.mol,
            analyzer.grids,
            analyzer.rdm1,
            settings,
            coeffs=coeffs,
            **kwargs
        )
        if orbs is not None:
            desc, deriv_arr = desc
            ddesc, eigvals = unpack_feature_derivs(orbs, labels, deriv_arr, en_list)
            return desc[None, :], ddesc, eigvals
        else:
            return desc[None, :]


def unpack_feature_derivs(orbs, labels, deriv_arr, en_list):
    ddesc = {}
    eigvals = {}
    for k in list(orbs.keys()):
        ddesc[k] = {}
        eigvals[k] = {}
    for n, (k, iorb) in enumerate(labels):
        ddesc[k][iorb] = deriv_arr[n].copy()
        eigvals[k][iorb] = en_list[n]
    return ddesc, eigvals


def unpack_eigvals(orbs, labels, en_list):
    eigvals = {}
    for k in list(orbs.keys()):
        eigvals[k] = {}
    for n, (k, iorb) in enumerate(labels):
        eigvals[k][iorb] = en_list[n]
    return eigvals


def _get_labels_and_coeffs(orbs, mo_coeff, mo_occ, mo_energy, spins=None):
    index_spin = spins is not None
    if index_spin:
        labels = ([], [])
        coeff_list = ([], [])
        en_list = ([], [])
    else:
        labels = []
        coeff_list = []
        en_list = []
    for k, v in orbs.items():
        if k == "O":
            inds = (mo_occ > 0.5).nonzero()[0][::-1]
        elif k == "U":
            inds = (mo_occ <= 0.5).nonzero()[0]
        else:
            inds = np.arange(mo_occ.size)
        mo_sub = mo_coeff[:, inds]
        en_sub = mo_energy[inds]
        if index_spin:
            spin_sub = spins[inds]
        else:
            spin_sub = None
        for iorb in v:
            l = (k, iorb)
            c = mo_sub[:, iorb]
            e = en_sub[iorb]
            if index_spin:
                s = spin_sub[iorb]
                labels[s].append(l)
                coeff_list[s].append(c)
                en_list[s].append(e)
            else:
                labels.append(l)
                coeff_list.append(c)
                en_list.append(e)
    return labels, coeff_list, en_list


def get_labels_and_coeffs(orbs, mo_coeff, mo_occ, mo_energy):
    spinpol = mo_occ.ndim == 2
    norb = mo_occ.shape[-1]
    sep_spins = False
    if spinpol:
        sep_spins = (isinstance(orbs, tuple) or isinstance(orbs, list)) and len(
            orbs
        ) == 2
        if sep_spins:
            labels_a, coeffs_a, en_list_a = _get_labels_and_coeffs(
                orbs[0], mo_coeff[0], mo_occ[0], mo_energy[0]
            )
            labels_b, coeffs_b, en_list_b = _get_labels_and_coeffs(
                orbs[1], mo_coeff[1], mo_occ[1], mo_energy[1]
            )
            labels = (labels_a, labels_b)
            coeffs = (coeffs_a, coeffs_b)
            en_list = (en_list_a, en_list_b)
        else:
            assert isinstance(orbs, dict)
            spins = np.append(
                np.zeros(norb, dtype=np.int32),
                np.ones(norb, dtype=np.int32),
            )
            mo_occ = mo_occ.flatten()
            mo_energy = mo_energy.flatten()
            mo_coeff = mo_coeff.transpose(1, 0, 2).reshape(norb, 2 * norb)
            inds = np.argsort(mo_energy)
            spins = np.ascontiguousarray(spins[inds])
            mo_occ = np.ascontiguousarray(mo_occ[inds])
            mo_energy = np.ascontiguousarray(mo_energy[inds])
            mo_coeff = np.ascontiguousarray(mo_coeff[:, inds])
            labels, coeffs, en_list = _get_labels_and_coeffs(
                orbs, mo_coeff, mo_occ, mo_energy, spins=spins
            )
    else:
        labels, coeffs, en_list = _get_labels_and_coeffs(
            orbs, mo_coeff, 0.5 * mo_occ, mo_energy
        )
    return labels, coeffs, en_list, sep_spins


def get_full_rho(ni, mol, dms, grids, xctype):
    hermi = 1
    max_memory = 2000
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_deriv = 1
    Nrhofeat = 4 if xctype == "GGA" else 5
    rho_full = np.zeros(
        (nset, Nrhofeat, grids.weights.size), dtype=np.float64, order="C"
    )
    ip0 = 0
    for ao, mask, weight, coords in ni.block_loop(
        mol, grids, nao, ao_deriv, max_memory
    ):
        mask = None  # TODO remove once screening is fixed
        ip1 = ip0 + weight.size
        for idm in range(nset):
            rho = make_rho(idm, ao, mask, xctype)
            rho_full[idm, :, ip0:ip1] = rho
        ip0 = ip1
    return rho_full


def get_full_rho_with_fl(ni, mol, dms, grids):
    xctype = "MGGA"
    hermi = 1
    max_memory = 2000
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_deriv = 1
    Nrhofeat = ni.plan.settings.nrho + 5
    rho_full = np.zeros(
        (nset, Nrhofeat, grids.weights.size), dtype=np.float64, order="C"
    )
    ip0 = 0
    for ao, mask, weight, coords in ni.block_loop(
        mol, grids, nao, ao_deriv, max_memory
    ):
        mask = None  # TODO remove once screening is fixed
        ip1 = ip0 + weight.size
        for idm in range(nset):
            rho = make_rho(idm, ao, mask, xctype)
            rho_full[idm, :, ip0:ip1] = rho
        ip0 = ip1
    return rho_full


def get_mo_densities(ni, mol, mo_vecs, grids, xctype):
    nmo = mo_vecs.shape[0]
    mo_coeff = np.ascontiguousarray(mo_vecs[:, :, np.newaxis])
    mo_occ = np.ones((nmo, 1))
    dms = np.empty((nmo, mol.nao_nr(), 0))
    dms = lib.tag_array(dms, mo_coeff=mo_coeff, mo_occ=mo_occ)
    return get_full_rho(ni, mol, dms, grids, xctype)


def get_mo_densities_with_fl(ni, mol, mo_vecs, grids):
    nmo = mo_vecs.shape[0]
    mo_coeff = np.ascontiguousarray(mo_vecs[:, :, np.newaxis])
    mo_occ = np.ones((nmo, 1))
    dms = np.empty((nmo, mol.nao_nr(), 0))
    dms = lib.tag_array(dms, mo_coeff=mo_coeff, mo_occ=mo_occ)
    return get_full_rho_with_fl(ni, mol, dms, grids)


def _density_getter(mol, pgrids, dms, nldf_settings, coeffs=None, **kwargs):
    ni = NumInt()
    desc = get_full_rho(ni, mol, dms, pgrids, "MGGA")[0]
    if coeffs is not None:
        if len(coeffs) == 0:
            return desc, 0
        coeffs = np.array(coeffs)
        ddesc = get_mo_densities(ni, mol, coeffs, pgrids, "MGGA")
        return desc, ddesc
    else:
        return desc


def _sl_desc_getter(mol, pgrids, dms, settings, coeffs=None, **kwargs):
    ni = NumInt()
    xctype = "MGGA"
    prho = get_full_rho(ni, mol, dms, pgrids, xctype)[0]
    plan = SemilocalPlan(settings, 1)
    if coeffs is None:
        return plan.get_feat(prho[None, :])[0]
    else:
        coeffs = np.array(coeffs)
        feat = plan.get_feat(prho[None, :])[0]
        dprho_dphi = get_mo_densities(ni, mol, coeffs, pgrids, xctype)
        occds = []
        nmo = coeffs.shape[0]
        for imo in range(nmo):
            occds.append(plan.get_occd(prho[None, :], dprho_dphi[imo : imo + 1])[1][0])
        return feat, np.stack(occds)


def _nldf_desc_getter(
    mol,
    pgrids,
    dms,
    nldf_settings,
    coeffs=None,
    inner_grids_level=3,
    inner_grids=None,
    **kwargs
):
    kwargs["lmax"] = kwargs.get("lmax") or DEFAULT_CIDER_LMAX
    kwargs["interpolator_type"] = "train_gen"
    if inner_grids is None:
        grids = CiderGrids(mol, lmax=kwargs["lmax"])
        grids.level = inner_grids_level
        grids.build(with_non0tab=False)
    else:
        grids = inner_grids
    nldf_generator = PyscfNLDFGenerator.from_mol_and_settings(
        mol, grids.grids_indexer, 1, nldf_settings, **kwargs
    )
    nldf_generator.interpolator.set_coords(pgrids.coords)
    ni = NumInt()
    xctype = "MGGA"

    rho = get_full_rho(ni, mol, dms, grids, xctype)[0]
    prho = get_full_rho(ni, mol, dms, pgrids, xctype)[0]
    if coeffs is not None and len(coeffs) > 0:
        coeffs = np.array(coeffs)
        drho_dphi = get_mo_densities(ni, mol, coeffs, grids, xctype)
        dprho_dphi = get_mo_densities(ni, mol, coeffs, pgrids, xctype)
    else:
        drho_dphi = np.empty(0)
        dprho_dphi = np.empty(0)

    res = nldf_generator.get_features_and_occ_derivs(rho, drho_dphi, prho, dprho_dphi)
    if coeffs is None:
        return res[0]
    elif len(coeffs) > 0:
        return res
    else:
        return res[0], 0


def _fl_desc_getter(mol, pgrids, dms, flsettings, coeffs=None, **kwargs):
    plan = FracLaplPlan(flsettings, 1)
    ni = FLNumInt(plan)
    desc = get_full_rho_with_fl(ni, mol, dms, pgrids)
    feat = plan.get_feat(desc, make_l1_data_copy=False)
    if coeffs is not None:
        if len(coeffs) == 0:
            return feat[0], 0
        coeffs = np.array(coeffs)
        ddesc = get_mo_densities_with_fl(ni, mol, coeffs, pgrids)
        dfeat = plan.get_occd(desc[0], ddesc)
        return feat[0], dfeat
    else:
        return feat[0]


def _sdmx_desc_getter(mol, pgrids, dm, settings, coeffs=None, **kwargs):
    exxgen = EXXSphGenerator.from_settings_and_mol(settings, 1, mol, **kwargs)
    ni = NumInt()
    ngrids = pgrids.coords.shape[0]
    nfeat = settings.nfeat
    desc = np.zeros((nfeat, ngrids))
    if coeffs is None:
        ddesc = 0
    else:
        coeffs = np.asarray(coeffs)
        ddesc = np.zeros((coeffs.shape[0], nfeat, ngrids))
    max_memory = 2000
    nao = mol.nao_nr()
    ip0 = 0
    maxmem = int(max_memory / exxgen.plan.nalpha) + 1
    if settings.n1terms > 0:
        maxmem //= 4
        maxmem += 1
    for ao, mask, weight, coords in ni.block_loop(mol, pgrids, nao, 0, maxmem):
        # TODO implement screening
        ip1 = ip0 + weight.size
        if coeffs is None or len(coeffs) == 0:
            desc[:, ip0:ip1] = exxgen.get_features(dm, mol, coords)
        else:
            desc[:, ip0:ip1], ddesc[:, :, ip0:ip1] = exxgen.get_feat_and_occd(
                dm, coeffs, mol, coords
            )
        ip0 = ip1
    if coeffs is None:
        return desc
    elif len(coeffs) == 0:
        return desc, 0
    return desc, ddesc



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#


import numpy as np
import yaml
from gpaw import mpi
from gpaw.xc.gga import GGA
from gpaw.xc.libvdwxc import FFTDistribution
from gpaw.xc.libxc import LibXC
from gpaw.xc.mgga import MGGA

from ciderpress.dft.density_util import get_sigma
from ciderpress.dft.plans import NLDFSplinePlan
from ciderpress.gpaw.nldf_interface import LibCiderPW

CIDERPW_GRAD_MODE_NONE = 0
CIDERPW_GRAD_MODE_FORCE = 1
CIDERPW_GRAD_MODE_STRESS = 2


class _CiderBase:

    is_cider_functional = True

    has_paw = False

    def __init__(
        self, cider_kernel, Nalpha=None, lambd=1.8, encut=300, world=None, **kwargs
    ):
        if world is None:
            self.world = mpi.world
        else:
            self.world = world

        self.cider_kernel = cider_kernel
        nldf_settings = cider_kernel.mlfunc.settings.nldf_settings
        if nldf_settings.is_empty:
            # TODO these numbers don't matter but must be defined, which is messy
            self.Nalpha = 10
            self.lambd = 2.0
            self.encut = 100
        else:
            if Nalpha is None:
                amax = encut
                # amin = np.min(consts[:, -1]) / np.e
                # TODO better choice here? depend on feat_params too?
                amin = nldf_settings.theta_params[0]
                if hasattr(nldf_settings, "feat_params"):
                    amin = min(amin, np.min(np.array(nldf_settings.feat_params)[:, 0]))
                amin /= 64
                Nalpha = int(np.ceil(np.log(amax / amin) / np.log(lambd))) + 1
                lambd = np.exp(np.log(amax / amin) / (Nalpha - 1))
            self.Nalpha = Nalpha
            self.lambd = lambd
            self.encut = encut

        self._plan = None
        self.distribution = None
        self.fft_obj = None
        self.c_asiq = None
        self._save_c_asiq = None
        self._save_v_avsiq = None
        self._global_redistributor = None
        self.aux_gd = None

        self.setup_name = "PBE"

    def get_mlfunc_data(self):
        return yaml.dump(self.cider_kernel.mlfunc, Dumper=yaml.CDumper)

    def get_setup_name(self):
        return self.setup_name

    def todict(self):
        kernel_params = self.cider_kernel.todict()
        xc_params = {
            "lambd": self.lambd,
            "encut": self.encut,
        }
        return {
            "kernel_params": kernel_params,
            "xc_params": xc_params,
        }

    def _check_parallelization(self, wfs):
        if wfs.world.size > self.gd.comm.size:
            import warnings

            from gpaw.utilities.grid import GridRedistributor

            self.aux_gd = self.gd.new_descriptor(comm=wfs.world)
            self._global_redistributor = GridRedistributor(
                wfs.world, wfs.kptband_comm, self.gd, self.aux_gd
            )
            self.distribution = FFTDistribution(
                self.aux_gd, [self.aux_gd.comm.size, 1, 1]
            )
            warnings.warn(
                "You are using Cider with only %d out of %d available cores. "
                "The internal Cider routines will redistribute the density "
                "so that all cores can be used. Setting augment_grids=True in "
                "the parallel option is recommended and might be more efficient."
                % (self.gd.comm.size, wfs.world.size)
            )
        else:
            self.aux_gd = self.gd
            self._global_redistributor = None

    def initialize(self, density, hamiltonian, wfs):
        if self.type == "GGA":
            GGA.initialize(self, density, hamiltonian, wfs)
        elif self.type == "MGGA":
            MGGA.initialize(self, density, hamiltonian, wfs)
        else:
            raise NotImplementedError
        self.timer = wfs.timer
        self.world = wfs.world
        self.dens = density
        self._plan = None
        self.aux_gd = None
        self._check_parallelization(wfs)

    def _setup_plan(self):
        nldf_settings = self.cider_kernel.mlfunc.settings.nldf_settings
        need_plan = self._plan is None or self._plan.nspin != self.nspin
        need_plan = need_plan and not nldf_settings.is_empty
        if need_plan:
            self._plan = NLDFSplinePlan(
                nldf_settings,
                self.nspin,
                self.encut / self.lambd ** (self.Nalpha - 1),
                self.lambd,
                self.Nalpha,
                coef_order="gq",
                alpha_formula="etb",
            )
        self.fft_obj = LibCiderPW(
            self.aux_gd.N_c, self.aux_gd.cell_cv, self.aux_gd.comm, plan=self._plan
        )

    def set_grid_descriptor(self, gd):
        self.gd = gd
        self.distribution = FFTDistribution(gd, [gd.comm.size, 1, 1])

    def set_fft_work(self, s, pot=False):
        pass

    def get_fft_work(self, s, pot=False, grad_mode=CIDERPW_GRAD_MODE_NONE):
        if grad_mode in [CIDERPW_GRAD_MODE_FORCE, CIDERPW_GRAD_MODE_STRESS]:
            self._save_v_avsiq = {}

    def _set_paw_terms(self):
        pass

    def _calculate_paw_energy_and_potential(self, grad_mode=CIDERPW_GRAD_MODE_NONE):
        if grad_mode == CIDERPW_GRAD_MODE_STRESS:
            return 0

    def _get_dH_asp(self):
        pass

    def call_fwd(self, rho_tuple, grad_mode=CIDERPW_GRAD_MODE_NONE):
        p, dp = None, None
        plan = self._plan
        self._set_paw_terms()
        arg_sg, darg_sg = plan.get_interpolation_arguments(rho_tuple, i=-1)
        fun_sg, dfun_sg = plan.get_function_to_convolve(rho_tuple)
        shape = (plan.nspin, plan.nldf_settings.nfeat) + arg_sg.shape[1:]
        feat_sig = np.empty(shape)
        shape = (plan.nspin, plan.num_vj) + arg_sg.shape[1:]
        dfeat_sjg = np.empty(shape)
        if grad_mode == CIDERPW_GRAD_MODE_STRESS:
            self._theta_s_gq = {}
        for s in range(plan.nspin):
            _rho_tuple = tuple(x[s] for x in rho_tuple)
            p, dp = plan.get_interpolation_coefficients(
                arg_sg[s], i=-1, vbuf=p, dbuf=dp
            )
            p.shape = fun_sg.shape[-3:] + (p.shape[-1],)
            self.fft_obj.set_work(fun_sg[s], p)
            self.set_fft_work(s, pot=False)
            if grad_mode == CIDERPW_GRAD_MODE_STRESS:
                self._theta_s_gq[s] = self.fft_obj.copy_work_array()
            self.timer.start("FFT and kernel")
            self.fft_obj.compute_forward_convolution()
            self.timer.stop("FFT and kernel")
            self.timer.start("Features")
            self.get_fft_work(s, pot=False, grad_mode=grad_mode)
            for i in range(plan.num_vj):
                a_g = plan.get_interpolation_arguments(_rho_tuple, i=i)[0]
                p, dp = plan.get_interpolation_coefficients(a_g, i=i, vbuf=p, dbuf=dp)
                if False:  # TODO might need this later
                    coeff_multipliers = None
                    p[:] *= coeff_multipliers
                    dp[:] *= coeff_multipliers
                if False:  # TODO might need this later
                    plan.get_transformed_interpolation_terms(
                        p, i=i, fwd=True, inplace=True
                    )
                    plan.get_transformed_interpolation_terms(
                        dp, i=i, fwd=True, inplace=True
                    )
                p.shape = a_g.shape + (plan.nalpha,)
                dp.shape = a_g.shape + (plan.nalpha,)
                self.fft_obj.fill_vj_feature_(feat_sig[s, i], p)
                self.fft_obj.fill_vj_feature_(dfeat_sjg[s, i], dp)
            self.timer.stop()
        return feat_sig, dfeat_sjg, arg_sg, darg_sg, fun_sg, dfun_sg, p, dp

    def call_bwd(
        self,
        rho_tuple,
        vfeat_sig,
        dfeat_sjg,
        arg_sg,
        fun_sg,
        v_sg,
        dedsigma_xg,
        dedtau_sg,
        p_gq=None,
        dp_gq=None,
        grad_mode=CIDERPW_GRAD_MODE_NONE,
    ):
        dedarg_sg = np.empty(v_sg.shape)
        dedfun_sg = np.empty(v_sg.shape)
        plan = self._plan
        paw_res = self._calculate_paw_energy_and_potential(grad_mode=grad_mode)
        if grad_mode == CIDERPW_GRAD_MODE_FORCE:
            fs = {}
            for a, v_vsiq in self._save_v_avsiq.items():
                fs[a] = np.einsum("vsiq,siq->v", v_vsiq, self.c_asiq[a])
        elif grad_mode == CIDERPW_GRAD_MODE_STRESS:
            fs = paw_res * np.eye(3)
            for a, v_vvsiq in self._save_v_avsiq.items():
                fs += np.einsum("uvsiq,siq->uv", v_vvsiq, self.c_asiq[a])
        else:
            fs = None
        for s in range(plan.nspin):
            self.timer.start("Feature potential")
            if grad_mode == CIDERPW_GRAD_MODE_STRESS:
                self.fft_obj.fftv2(self._theta_s_gq[s])
            self.fft_obj.reset_work()
            for i in range(plan.num_vj):
                _rho_tuple = (x[s] for x in rho_tuple)
                a_g, da_g = plan.get_interpolation_arguments(_rho_tuple, i=i)
                p_gq, dp_gq = plan.get_interpolation_coefficients(
                    a_g, i=i, vbuf=p_gq, dbuf=dp_gq
                )
                p_gq.shape = a_g.shape + (plan.nalpha,)
                dp_gq.shape = a_g.shape + (plan.nalpha,)
                self.fft_obj.fill_vj_potential_(vfeat_sig[s, i], p_gq)
                tmp = dfeat_sjg[s, i] * vfeat_sig[s, i]
                v_sg[s] += da_g[0] * tmp
                dedsigma_xg[2 * s] += da_g[1] * tmp
                if len(rho_tuple) == 3:
                    dedtau_sg[s] += da_g[2] * tmp
                # dedrho_sxg[s, 0] += da_g[0] * tmp
                # dedrho_sxg[s, 1:4] += 2 * da_g[1] * rho_sxg[s, 1:4] * tmp
                # if tau_sg is not None:
                #    dedrho_sxg[s, 4] += da_g[2] * tmp
            self.timer.stop()
            self.set_fft_work(s, pot=True)
            self.timer.start("FFT and kernel")
            if grad_mode == CIDERPW_GRAD_MODE_STRESS:
                fs += self.fft_obj.compute_backward_convolution_with_stress(
                    self._theta_s_gq[s]
                )
            else:
                self.fft_obj.compute_backward_convolution()
            self.timer.stop("FFT and kernel")
            self.get_fft_work(s, pot=True, grad_mode=grad_mode)
            p_gq, dp_gq = plan.get_interpolation_coefficients(
                arg_sg[s], i=-1, vbuf=p_gq, dbuf=dp_gq
            )
            p_gq.shape = fun_sg.shape[-3:] + (p_gq.shape[-1],)
            dp_gq.shape = fun_sg.shape[-3:] + (dp_gq.shape[-1],)
            self.fft_obj.fill_vj_feature_(dedarg_sg[s], dp_gq)
            self.fft_obj.fill_vj_feature_(dedfun_sg[s], p_gq)
        if grad_mode == CIDERPW_GRAD_MODE_FORCE:
            for a, v_vsiq in self._save_v_avsiq.items():
                fs[a] += np.einsum("vsiq,siq->v", v_vsiq, self._save_c_asiq[a])
        elif grad_mode == CIDERPW_GRAD_MODE_STRESS:
            for a, v_vvsiq in self._save_v_avsiq.items():
                fs += np.einsum("uvsiq,siq->uv", v_vvsiq, self._save_c_asiq[a])
        dedarg_sg[:] *= fun_sg
        self._get_dH_asp()
        if fs is not None:
            return dedarg_sg, dedfun_sg, fs
        else:
            return dedarg_sg, dedfun_sg

    def initialize_more_things(self):
        pass

    @property
    def is_initialized(self):
        if self._plan is None or self._plan.nspin != self.nspin:
            return False
        else:
            return True

    def calc_cider(
        self,
        e_g,
        rho_sxg,
        dedrho_sxg,
        grad_mode=CIDERPW_GRAD_MODE_NONE,
    ):
        nspin = len(rho_sxg)
        self.nspin = nspin
        if not self.is_initialized:
            self._setup_plan()
            self.fft_obj.initialize_backend()
            self.initialize_more_things()
        plan = self._plan
        sigma_xg = get_sigma(rho_sxg[:, 1:4])
        dedsigma_xg = np.zeros_like(sigma_xg)
        rho_tuple = plan.get_rho_tuple(rho_sxg, with_spin=True)
        nt_sg = rho_tuple[0]
        v_sg = np.zeros_like(nt_sg)
        if len(rho_tuple) == 2:
            tau_sg = None
            dedtau_sg = None
        else:
            tau_sg = rho_tuple[2].copy()
            dedtau_sg = np.zeros_like(tau_sg)

        res = self.call_fwd(rho_tuple, grad_mode=grad_mode)
        feat_sig, dfeat_sjg, arg_sg, darg_sg, fun_sg, dfun_sg, p_gq, dp_gq = res
        # TODO version i features
        feat_sig[:] *= plan.nspin
        self.timer.start("eval xc")
        if e_g.size > 0:
            if tau_sg is None:  # GGA
                vfeat_sig = self.cider_kernel.calculate(
                    e_g, nt_sg, v_sg, sigma_xg, dedsigma_xg, feat_sig
                )
            else:  # MGGA
                vfeat_sig = self.cider_kernel.calculate(
                    e_g, nt_sg, v_sg, sigma_xg, dedsigma_xg, tau_sg, dedtau_sg, feat_sig
                )
        else:
            vfeat_sig = np.zeros_like(feat_sig)
        self.timer.stop()

        vfeat_sig[:] *= plan.nspin
        res = self.call_bwd(
            rho_tuple,
            vfeat_sig,
            dfeat_sjg,
            arg_sg,
            fun_sg,
            v_sg,
            dedsigma_xg,
            dedtau_sg,
            p_gq=p_gq,
            dp_gq=dp_gq,
            grad_mode=grad_mode,
        )
        if grad_mode in [CIDERPW_GRAD_MODE_FORCE, CIDERPW_GRAD_MODE_STRESS]:
            dedarg_sg, dedfun_sg, fs = res
        else:
            dedarg_sg, dedfun_sg = res
            fs = None
        self.timer.start("add potential")
        for i, term in enumerate(darg_sg):
            term[:] *= dedarg_sg
            term[:] += dedfun_sg * dfun_sg[i]
        for s in range(plan.nspin):
            v_sg[s] += darg_sg[0][s]
            dedsigma_xg[2 * s] += darg_sg[1][s]
            if tau_sg is not None:
                dedtau_sg[s] += darg_sg[2][s]

        dedrho_sxg[:, 0] += v_sg
        dedrho_sxg[:, 1:4] += 2 * rho_sxg[:, 1:4] * dedsigma_xg[::2, None, ...]
        if plan.nspin == 2:
            dedrho_sxg[:, 1:4] += rho_sxg[::-1, 1:4] * dedsigma_xg[1:2, None, ...]
        if tau_sg is not None:
            dedrho_sxg[:, 4] += dedtau_sg
        self.timer.stop()
        return fs

    def _distribute_to_cider_grid(self, n_xg):
        if self._global_redistributor is not None:
            n_xg = self._global_redistributor.distribute(n_xg)
        shape = n_xg.shape[:-3]
        ndist_xg = self.distribution.block_zeros(shape)
        self.distribution.gd2block(n_xg, ndist_xg)
        return ndist_xg

    def _add_from_cider_grid(self, d_xg, ddist_xg):
        if self._global_redistributor is not None:
            shape = d_xg.shape[:-3]
            tmp_xg = self._global_redistributor.aux_gd.zeros(shape)
            self.distribution.block2gd_add(ddist_xg, tmp_xg)
            d_xg[:] += self._global_redistributor.collect(tmp_xg)
        else:
            self.distribution.block2gd_add(ddist_xg, d_xg)

    def _get_taut(self, n_sg):
        if self.type == "MGGA":
            if self.fixed_ke:
                taut_sG = self._fixed_taut_sG
                if not self._taut_gradv_init:
                    self._taut_gradv_init = True
                    # ensure initialization for calculating potential
                    self.wfs.calculate_kinetic_energy_density()
            else:
                taut_sG = self.wfs.calculate_kinetic_energy_density()

            if taut_sG is None:
                taut_sG = self.wfs.gd.zeros(len(n_sg))

            taut_sg = np.empty_like(n_sg)

            for taut_G, taut_g in zip(taut_sG, taut_sg):
                self.distribute_and_interpolate(taut_G, taut_g)

            dedtaut_sg = np.zeros_like(n_sg)
        else:
            taut_sg = None
            dedtaut_sg = None
            taut_sG = None
        return taut_sg, dedtaut_sg, taut_sG

    def _get_cider_inputs(self, n_sg, taut_sg):
        rho_sxg = get_rho_sxg(n_sg, self.grad_v, taut_sg)
        shape = rho_sxg.shape[:2]
        rho_sxg.shape = (rho_sxg.shape[0] * rho_sxg.shape[1],) + rho_sxg.shape[-3:]
        rho_sxg = self._distribute_to_cider_grid(rho_sxg)
        rho_sxg.shape = shape + rho_sxg.shape[-3:]
        self.nspin = len(rho_sxg)
        dedrho_sxg = np.zeros_like(rho_sxg)
        _e = self._distribute_to_cider_grid(np.zeros_like(n_sg[0])[None, :])[0]
        return _e, rho_sxg, dedrho_sxg

    def calculate_impl(self, gd, n_sg, v_sg, e_g):
        taut_sg, dedtaut_sg, taut_sG = self._get_taut(n_sg)
        _e, rho_sxg, dedrho_sxg = self._get_cider_inputs(n_sg, taut_sg)
        tmp_dedrho_sxg = np.zeros((len(n_sg), rho_sxg.shape[1]) + e_g.shape)
        e_g[:] = 0.0
        self.calc_cider(_e, rho_sxg, dedrho_sxg)
        self._add_from_cider_grid(e_g[None, :], _e[None, :])
        for s in range(len(n_sg)):
            self._add_from_cider_grid(tmp_dedrho_sxg[s], dedrho_sxg[s])
        v_sg[:] = tmp_dedrho_sxg[:, 0]
        add_gradient_correction(self.grad_v, tmp_dedrho_sxg, v_sg)

        if dedtaut_sg is not None:
            dedtaut_sg[:] = tmp_dedrho_sxg[:, 4]
            self.dedtaut_sG = self.wfs.gd.empty(self.wfs.nspins)
            self.ekin = 0.0
            for s in range(self.wfs.nspins):
                self.restrict_and_collect(dedtaut_sg[s], self.dedtaut_sG[s])
                self.ekin -= self.wfs.gd.integrate(self.dedtaut_sG[s] * taut_sG[s])


class CiderGGA(_CiderBase, GGA):
    def __init__(self, cider_kernel, **kwargs):
        sl_settings = cider_kernel.mlfunc.settings.sl_settings
        if not sl_settings.is_empty and sl_settings.level != "GGA":
            raise ValueError("CiderGGA only supports GGA functionals!")
        _CiderBase.__init__(self, cider_kernel, **kwargs)

        GGA.__init__(self, LibXC("PBE"), stencil=2)
        self.type = "GGA"
        self.name = "CIDER_{}".format(
            self.cider_kernel.xmix
        )  # TODO more informative name

    def todict(self):
        d = super(CiderGGA, self).todict()
        d["_cider_type"] = "CiderGGA"
        return d

    def set_grid_descriptor(self, gd):
        GGA.set_grid_descriptor(self, gd)
        _CiderBase.set_grid_descriptor(self, gd)

    def set_positions(self, spos_ac):
        self.spos_ac = spos_ac

    def stress_tensor_contribution(self, n_sg):
        e_g = np.zeros(n_sg.shape[1:])
        stress1_vv = np.zeros((3, 3))
        _e, rho_sxg, dedrho_sxg = self._get_cider_inputs(n_sg, None)
        tmp_dedrho_sxg = np.zeros((len(n_sg), rho_sxg.shape[1]) + e_g.shape)
        e_g[:] = 0.0
        fs = self.calc_cider(
            _e, rho_sxg, dedrho_sxg, grad_mode=CIDERPW_GRAD_MODE_STRESS
        )
        stress1_vv[:] += 0.5 * (fs + fs.T)
        self.world.sum(stress1_vv, 0)

        self._add_from_cider_grid(e_g[None, :], _e[None, :])
        for s in range(len(n_sg)):
            self._add_from_cider_grid(tmp_dedrho_sxg[s], dedrho_sxg[s])
        dedrho_sxg = tmp_dedrho_sxg
        tmp_rho_sxg = np.zeros_like(tmp_dedrho_sxg)
        for s in range(len(n_sg)):
            self._add_from_cider_grid(tmp_rho_sxg[s], rho_sxg[s])
        rho_sxg = tmp_rho_sxg

        def integrate(a1_g, a2_g=None):
            return self.gd.integrate(a1_g, a2_g, global_integral=False)

        P = integrate(e_g)
        nspin = n_sg.shape[0]
        for s in range(nspin):
            for x in range(rho_sxg.shape[1]):
                P -= integrate(dedrho_sxg[s, x], rho_sxg[s, x])

        stress_vv = P * np.eye(3)
        for s in range(nspin):
            for v1 in range(3):
                for v2 in range(3):
                    stress_vv[v1, v2] -= integrate(
                        rho_sxg[s, v1 + 1], dedrho_sxg[s, v2 + 1]
                    )

        stress_vv = 0.5 * (stress_vv + stress_vv.T)

        self.gd.comm.sum(stress_vv)

        stress_vv += stress1_vv

        self.gd.comm.broadcast(stress_vv, 0)
        return stress_vv


def get_rho_sxg(n_sg, grad_v, tau_sg=None):
    nspin = len(n_sg)
    shape = list(n_sg.shape)
    if tau_sg is None:
        shape = shape[:1] + [4] + shape[1:]
    else:
        shape = shape[:1] + [5] + shape[1:]
    rho_sxg = np.empty(shape)
    rho_sxg[:, 0] = n_sg
    for v in range(3):
        for s in range(nspin):
            grad_v[v](n_sg[s], rho_sxg[s, v + 1])
    if tau_sg is not None:
        rho_sxg[:, 4] = tau_sg
    return rho_sxg


def add_gradient_correction(grad_v, dedrho_sxg, v_sg):
    nspin = len(v_sg)
    vv_g = np.empty_like(v_sg[0])
    for v in range(3):
        for s in range(nspin):
            grad_v[v](dedrho_sxg[s, v + 1], vv_g)
            v_sg[s] -= vv_g


class CiderMGGA(_CiderBase, MGGA):
    def __init__(self, cider_kernel, **kwargs):
        sl_settings = cider_kernel.mlfunc.settings.sl_settings
        if (not sl_settings.is_empty) and sl_settings.level != "MGGA":
            raise ValueError("CiderMGGA only supports MGGA functionals!")
        _CiderBase.__init__(self, cider_kernel, **kwargs)

        MGGA.__init__(self, LibXC("PBE"), stencil=2)
        self.type = "MGGA"
        self.name = "CIDER_{}".format(self.cider_kernel.xmix)

    def todict(self):
        d = super(CiderMGGA, self).todict()
        d["_cider_type"] = "CiderMGGA"
        return d

    def set_positions(self, spos_ac):
        self.spos_ac = spos_ac

    def set_grid_descriptor(self, gd):
        MGGA.set_grid_descriptor(self, gd)
        _CiderBase.set_grid_descriptor(self, gd)

    def calculate_paw_correction(
        self, setup, D_sp, dEdD_sp=None, addcoredensity=True, a=None
    ):
        return 0

    def add_forces(self, F_av):
        raise NotImplementedError(
            "Forces not implemented for CIDER MGGA with NCPP. "
            "Please use PAW setups to get forces."
        )

    def stress_tensor_contribution(self, n_sg):
        raise NotImplementedError("MGGA stress requires PAW setups")



import os

import numpy as np

DATA_PATH = os.path.dirname(os.path.abspath(__file__))


def get_hgbs_min_exps(Z):
    dpath = os.path.join(DATA_PATH, "expnt_mins.npy")
    return np.load(dpath)[int(Z)]


def get_hgbs_max_exps(Z):
    dpath = os.path.join(DATA_PATH, "expnt_maxs.npy")
    return np.load(dpath)[int(Z)]



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np

from ciderpress.models.kernels import (
    DiffAntisymRBF,
    DiffConstantKernel,
    SubsetARBF,
    SubsetRBF,
)


def get_rbf_kernel(
    indexes, length_scale, scale=1.0, opt_hparams=False, min_lscale=None
):
    if min_lscale is None:
        min_lscale = 0.01
    length_scale_bounds = (min_lscale, 10) if opt_hparams else "fixed"
    scale_bounds = (1e-5, 1e3) if opt_hparams else "fixed"
    return DiffConstantKernel(scale, constant_value_bounds=scale_bounds) * SubsetRBF(
        indexes,
        length_scale=length_scale[indexes],
        length_scale_bounds=length_scale_bounds,
    )


def get_antisym_rbf_kernel(length_scale, scale=1.0, opt_hparams=False, min_lscale=None):
    if min_lscale is None:
        min_lscale = 0.01
    length_scale_bounds = (min_lscale, 10) if opt_hparams else "fixed"
    scale_bounds = (1e-5, 1e3) if opt_hparams else "fixed"
    length_scale = np.append(
        0.5 * (length_scale[0] + length_scale[1]), length_scale[2:]
    )
    return DiffConstantKernel(
        scale, constant_value_bounds=scale_bounds
    ) * DiffAntisymRBF(
        length_scale=length_scale,
        length_scale_bounds=length_scale_bounds,
    )


def get_agpr_kernel(
    sinds,
    ainds,
    length_scale,
    scale=None,
    order=2,
    nsingle=1,
    opt_hparams=False,
    min_lscale=None,
):
    print(sinds, ainds, length_scale[sinds], length_scale[ainds])
    if min_lscale is None:
        min_lscale = 0.01
    length_scale_bounds = (min_lscale, 10) if opt_hparams else "fixed"
    scale_bounds = (1e-5, 1e5) if opt_hparams else "fixed"
    if scale is None:
        scale = [1.0] * (order + 1)
    if nsingle == 0:
        singles = None
    else:
        singles = SubsetRBF(
            sinds,
            length_scale=length_scale[sinds],
            length_scale_bounds=length_scale_bounds,
        )
    cov_kernel = SubsetARBF(
        ainds,
        order=order,
        length_scale=length_scale[ainds],
        scale=scale,
        length_scale_bounds=length_scale_bounds,
        scale_bounds=scale_bounds,
    )
    if singles is None:
        return cov_kernel
    else:
        return singles * cov_kernel



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np
from pyscf import lib
from pyscf.dft import numint
from pyscf.dft.gen_grid import ALIGNMENT_UNIT, make_mask
from pyscf.dft.numint import (
    OCCDROP,
    _contract_rho,
    _contract_rho_sparse,
    _dot_ao_ao_sparse,
    _dot_ao_dm,
    _dot_ao_dm_sparse,
    _empty_aligned,
    _NumIntMixin,
    _scale_ao_sparse,
    _sparse_enough,
)
from pyscf.gto.eval_gto import (
    BLKSIZE,
    CUTOFF,
    NBINS,
    _get_intor_and_comp,
    libcgto,
    make_loc,
    make_screen_index,
)
from pyscf.gto.mole import ANG_OF
from scipy.special import gamma, hyp1f1

from ciderpress.lib import load_library as load_cider_library

libcider = load_cider_library("libmcider")

GLOBAL_FRAC_LAPL_DATA = {}
DEFAULT_FRAC_LAPL_POWER = 0.5
DEFAULT_FRAC_LAPL_A = 0.25
DEFAULT_FRAC_LAPL_D = 0.002
DEFAULT_FRAC_LAPL_SIZE = 4000
XCUT_1F1 = 100


class FracLaplBuf:
    def __init__(self, a, d, size, lmax, s):
        print("INITIALIZING FLAPL", s)
        self.a = a
        self.d = d
        self.size = size
        self.lmax = lmax
        self.s = s
        self.buf = np.empty((lmax + 1, 4, size))
        x = a * (np.exp(d * np.arange(size)) - 1)
        f = np.zeros((lmax + 1, size))
        xbig = x[x >= XCUT_1F1]
        xmax = x[x < XCUT_1F1][-1]
        for l in range(lmax + 1):
            f[l] = hyp1f1(1.5 + s + l, 1.5 + l, -x)
            f[l] *= 2 ** (2 * s) * gamma(1.5 + s + l) / gamma(1.5 + l)
            fsmall = f[l, x < XCUT_1F1][-1]
            f[l, x > XCUT_1F1] = fsmall * (xmax / xbig) ** (1.5 + s + l)
        libcider.initialize_spline_1f1(
            self.buf.ctypes.data_as(ctypes.c_void_p),
            f.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(self.size),
            ctypes.c_int(self.lmax),
        )
        df = np.zeros((lmax + 1, size))
        for l in range(lmax + 1):
            df[l] = hyp1f1(2.5 + s + l, 2.5 + l, -x)
            df[l] *= 2 ** (2 * s) * gamma(2.5 + s + l) / gamma(2.5 + l)
            fsmall = df[l, x < XCUT_1F1][-1]
            df[l, x > XCUT_1F1] = fsmall * (xmax / xbig) ** (2.5 + s + l)
        self.dbuf = np.empty((lmax + 1, 4, size))
        libcider.initialize_spline_1f1(
            self.dbuf.ctypes.data_as(ctypes.c_void_p),
            df.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(self.size),
            ctypes.c_int(self.lmax),
        )

    def set_data(self):
        # libcider.set_spline_1f1(
        #    self.buf.ctypes.data_as(ctypes.c_void_p),
        #    ctypes.c_double(self.a),
        #    ctypes.c_double(self.d),
        #    ctypes.c_int(self.size),
        #    ctypes.c_int(self.lmax),
        #    ctypes.c_double(self.s),
        # )
        libcider.set_spline_1f1_with_grad(
            self.buf.ctypes.data_as(ctypes.c_void_p),
            self.dbuf.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_double(self.a),
            ctypes.c_double(self.d),
            ctypes.c_int(self.size),
            ctypes.c_int(self.lmax),
            ctypes.c_double(self.s),
        )


def _set_flapl(a=0.025, d=0.004, size=6000, lmax=6, s=0.5):
    s_key = int(s * 100)
    if s_key not in GLOBAL_FRAC_LAPL_DATA or GLOBAL_FRAC_LAPL_DATA[s_key].lmax < lmax:
        GLOBAL_FRAC_LAPL_DATA[s_key] = FracLaplBuf(a, d, size, lmax, s)
    GLOBAL_FRAC_LAPL_DATA[s_key].set_data()


def set_frac_lapl(s, lmax=6):
    _set_flapl(lmax=lmax, s=s)


def eval_flapl_gto(
    slst,
    mol,
    coords,
    shls_slice=None,
    non0tab=None,
    ao_loc=None,
    cutoff=None,
    out=None,
    debug=False,
    deriv=0,
):
    if non0tab is not None:
        if (non0tab == 0).any():
            # TODO implement some sort of screening later
            raise NotImplementedError
    if mol.cart:
        feval = "GTOval_cart_deriv%d" % deriv
    else:
        feval = "GTOval_sph_deriv%d" % deriv
    eval_name, comp = _get_intor_and_comp(mol, feval)
    comp = len(slst)

    atm = np.asarray(mol._atm, dtype=np.int32, order="C")
    bas = np.asarray(mol._bas, dtype=np.int32, order="C")
    env = np.asarray(mol._env, dtype=np.double, order="C")
    coords = np.asarray(coords, dtype=np.double, order="F")
    natm = atm.shape[0]
    nbas = bas.shape[0]
    ngrids = coords.shape[0]

    if eval_name == "GTOval_sph_deriv0":
        deriv = 0
    elif eval_name == "GTOval_sph_deriv1":
        deriv = 1
    else:
        raise NotImplementedError
    if debug:
        assert deriv == 0
        contract_fn = getattr(libcgto, "GTOcontract_exp0")
        eval_fn = getattr(libcgto, "GTOshell_eval_grid_cart")
    else:
        if deriv == 0:
            contract_fn = getattr(libcider, "GTOcontract_flapl0")
            eval_fn = getattr(libcgto, "GTOshell_eval_grid_cart")
        elif deriv == 1:
            contract_fn = getattr(libcider, "GTOcontract_flapl1")
            eval_fn = getattr(libcgto, "GTOshell_eval_grid_cart_deriv1")
            comp *= 4
        else:
            raise ValueError

    if ao_loc is None:
        ao_loc = make_loc(bas, eval_name)

    if shls_slice is None:
        shls_slice = (0, nbas)
    sh0, sh1 = shls_slice
    nao = ao_loc[sh1] - ao_loc[sh0]
    if "spinor" in eval_name:
        ao = np.ndarray(
            (2, comp, nao, ngrids), dtype=np.complex128, buffer=out
        ).transpose(0, 1, 3, 2)
    else:
        ao = np.ndarray((comp, nao, ngrids), buffer=out).transpose(0, 2, 1)

    if non0tab is None:
        if cutoff is None:
            non0tab = np.ones(((ngrids + BLKSIZE - 1) // BLKSIZE, nbas), dtype=np.uint8)
        else:
            non0tab = make_screen_index(mol, coords, shls_slice, cutoff)

    # drv = getattr(libcgto, eval_name)
    # normal call tree is GTOval_sph_deriv0 -> GTOval_sph -> GTOeval_sph_drv
    drv = getattr(libcgto, "GTOeval_sph_drv")
    for i, s in enumerate(slst):
        set_frac_lapl(s, lmax=mol._bas[:, ANG_OF].max())
        if deriv == 0:
            j = i
            param = (1, 1)
        else:
            j = i * 4
            param = (1, 4)
        drv(
            eval_fn,
            contract_fn,
            ctypes.c_double(1),
            ctypes.c_int(ngrids),
            (ctypes.c_int * 2)(*param),
            (ctypes.c_int * 2)(*shls_slice),
            ao_loc.ctypes.data_as(ctypes.c_void_p),
            ao[j].ctypes.data_as(ctypes.c_void_p),
            coords.ctypes.data_as(ctypes.c_void_p),
            non0tab.ctypes.data_as(ctypes.c_void_p),
            atm.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(natm),
            bas.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(nbas),
            env.ctypes.data_as(ctypes.c_void_p),
        )
    if comp == 1:
        if "spinor" in eval_name:
            ao = ao[:, 0]
        else:
            ao = ao[0]
    return ao


def eval_rho(
    mol,
    all_ao,
    dm,
    flsettings,
    non0tab=None,
    xctype="LDA",
    hermi=0,
    with_lapl=True,
    verbose=None,
):
    ao, kao = all_ao
    xctype = xctype.upper()
    ngrids, nao = ao.shape[-2:]

    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()
    if with_lapl:
        raise NotImplementedError
    else:
        rho = np.empty((5 + flsettings.nrho, ngrids))
        tau_idx = 4

    c0 = _dot_ao_dm(mol, ao[0], dm, non0tab, shls_slice, ao_loc)
    # rho[0] = numpy.einsum('pi,pi->p', ao[0], c0)
    rho[0] = _contract_rho(ao[0], c0)
    nk0 = flsettings.nk0
    nk1 = flsettings.nk1
    nd1 = flsettings.nd1

    for ik in range(nk0):
        ir = ik + 1 + tau_idx
        indk = flsettings.get_ik0(ik)
        rho[ir] = _contract_rho(kao[indk], c0)
        if not hermi:
            rho[ir] += _contract_rho(c0, kao[indk])
            rho[ir] *= 0.5

    for ik in range(nd1):
        ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * ik)
        indk = flsettings.get_ik0(ik)
        for i in range(3):
            rho[ir + i] = _contract_rho(c0, kao[indk + 1 + i])

    rho[tau_idx] = 0
    for ik in range(flsettings.ndd):
        ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
        rho[ir] = 0
    for i in range(1, 4):
        c1 = _dot_ao_dm(mol, ao[i], dm, non0tab, shls_slice, ao_loc)
        # rho[tau_idx] += numpy.einsum('pi,pi->p', c1, ao[i])
        rho[tau_idx] += _contract_rho(ao[i], c1)

        # rho[i] = numpy.einsum('pi,pi->p', c0, ao[i])
        rho[i] = _contract_rho(ao[i], c0)
        if hermi:
            rho[i] *= 2
        else:
            rho[i] += _contract_rho(c1, ao[0])

        for ik in range(flsettings.nk1):
            ir = 3 * ik + i + tau_idx + nk0
            indk = flsettings.get_ik0(ik)
            rho[ir] = _contract_rho(kao[indk], c1)
            if not hermi:
                rho[ir] += _contract_rho(c1, kao[indk])
                rho[ir] *= 0.5

        for ik in range(flsettings.ndd):
            ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
            indk = flsettings.get_ik0(ik) + i
            rho[ir] += _contract_rho(kao[indk], c1)
            if not hermi:
                rho[ir] += _contract_rho(c1, kao[indk])

    # tau = 1/2 (\nabla f)^2
    rho[tau_idx] *= 0.5
    for ik in range(flsettings.ndd):
        ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
        if not hermi:
            rho[ir] *= 0.5
    return rho


def eval_rho1(
    mol,
    all_ao,
    dm,
    flsettings,
    screen_index=None,
    xctype="LDA",
    hermi=0,
    with_lapl=True,
    cutoff=None,
    ao_cutoff=CUTOFF,
    pair_mask=None,
    verbose=None,
):
    ao, kao = all_ao
    if flsettings.nd1 != 0 or flsettings.ndd != 0:
        raise NotImplementedError
    xctype = xctype.upper()
    ngrids = ao.shape[-2]

    if cutoff is None:
        cutoff = CUTOFF
    cutoff = min(cutoff, 0.1)
    nbins = NBINS * 2 - int(NBINS * np.log(cutoff) / np.log(ao_cutoff))

    if pair_mask is None:
        ovlp_cond = mol.get_overlap_cond()
        pair_mask = ovlp_cond < -np.log(cutoff)

    ao_loc = mol.ao_loc_nr()
    if with_lapl:
        raise NotImplementedError
    else:
        rho = np.empty((5 + flsettings.nrho, ngrids))
        tau_idx = 4
    c0 = _dot_ao_dm_sparse(ao[0], dm, nbins, screen_index, pair_mask, ao_loc)
    rho[0] = _contract_rho_sparse(ao[0], c0, screen_index, ao_loc)
    nk0 = flsettings.nk0
    nk1 = flsettings.nk1
    nd1 = flsettings.nd1

    for ik in range(nk0):
        # TODO might be able to use some kind of screening here
        ir = ik + 1 + tau_idx
        indk = flsettings.get_ik0(ik)
        # rho[ir] = _contract_rho_sparse(kao[indk], c0, screen_index, ao_loc)
        rho[ir] = _contract_rho(kao[indk], c0)
        if not hermi:
            # rho[ir] += _contract_rho_sparse(c0, kao[indk], screen_index, ao_loc)
            rho[ir] += _contract_rho(c0, kao[indk])
            rho[ir] *= 0.5

    for ik in range(nd1):
        ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * ik)
        indk = flsettings.get_ik0(ik)
        for i in range(3):
            # rho[ir + i] = _contract_rho_sparse(
            #    c0, kao[indk + 1 + i], screen_index, ao_loc
            # )
            rho[ir + i] = _contract_rho(c0, kao[indk + 1 + i])

    rho[tau_idx] = 0
    for ik in range(flsettings.ndd):
        ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
        rho[ir] = 0
    for i in range(1, 4):
        c1 = _dot_ao_dm_sparse(ao[i], dm.T, nbins, screen_index, pair_mask, ao_loc)
        rho[tau_idx] += _contract_rho_sparse(ao[i], c1, screen_index, ao_loc)

        rho[i] = _contract_rho_sparse(ao[i], c0, screen_index, ao_loc)
        if hermi:
            rho[i] *= 2
        else:
            rho[i] += _contract_rho_sparse(c1, ao[0], screen_index, ao_loc)

        for ik in range(flsettings.nk1):
            ir = 3 * ik + i + tau_idx + nk0
            indk = flsettings.get_ik0(ik)
            # rho[ir] = _contract_rho_sparse(kao[indk], c1, screen_index, ao_loc)
            rho[ir] = _contract_rho(kao[indk], c1)
            if not hermi:
                # rho[ir] += _contract_rho_sparse(c1, kao[indk], screen_index, ao_loc)
                rho[ir] += _contract_rho(c1, kao[indk])
                rho[ir] *= 0.5

        for ik in range(flsettings.ndd):
            ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
            indk = flsettings.get_ik0(ik) + i
            # rho[ir] += _contract_rho_sparse(kao[indk], c1, screen_index, ao_loc)
            rho[ir] += _contract_rho(kao[indk], c1)
            if not hermi:
                # rho[ir] += _contract_rho_sparse(c1, kao[indk], screen_index, ao_loc)
                rho[ir] += _contract_rho(c1, kao[indk])

    rho[tau_idx] *= 0.5
    for ik in range(flsettings.ndd):
        ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
        if not hermi:
            rho[ir] *= 0.5
    return rho


def _odp_dot_sparse_(
    all_ao,
    wv,
    nbins,
    mask,
    pair_mask,
    ao_loc,
    flsettings,
    aow1=None,
    aow2=None,
    vmat=None,
    v1=None,
):
    ao, kao = all_ao
    wv[0] *= 0.5  # *.5 for v+v.conj().T
    wv[4] *= 0.5  # *.5 for 1/2 in tau
    if flsettings.nrho > 0:
        wv[5:] *= 0.5

    aow1 = _scale_ao_sparse(ao[:4], wv[:4], mask, ao_loc, out=aow1)
    nk0 = flsettings.nk0
    nk1 = flsettings.nk1
    nd1 = flsettings.nd1
    ndd = flsettings.ndd
    tau_idx = 4
    if nd1 > 0 or ndd > 0:
        wvtmp = np.zeros(kao.shape[:2])
        for ik in range(nk0):
            ir = ik + 1 + tau_idx
            indk = flsettings.get_ik0(ik)
            wvtmp[indk] = wv[ir]
            # rho[ir] = _contract_rho(kao[indk], c0)
            # if not hermi:
            #    rho[ir] += _contract_rho(c0, kao[indk])
            #    rho[ir] *= 0.5
        for ik in range(nd1):
            ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * ik)
            indk = flsettings.get_ik0(ik)
            wvtmp[indk + 1 : indk + 4] = wv[ir : ir + 3]
        aow2 = _scale_ao_sparse(kao, wvtmp, None, ao_loc, out=aow2)
        aow1 += aow2
        vmat = _dot_ao_ao_sparse(
            ao[0], aow1, None, nbins, None, pair_mask, ao_loc, hermi=0, out=vmat
        )
    elif nk0 > 0:
        # TODO might be able to use some kind of screening here
        # aow2 = _scale_ao_sparse(kao, wv[5:5 + nk0], mask, ao_loc, out=aow2)
        # aow1 += aow2
        # vmat = _dot_ao_ao_sparse(ao[0], aow1, None, nbins, mask, pair_mask,
        #                         ao_loc, hermi=0, out=vmat)
        aow2 = _scale_ao_sparse(kao, wv[5 : 5 + nk0], None, ao_loc, out=aow2)
        print(np.sum(aow2), np.sum(aow1))
        aow1 += aow2
        vmat = _dot_ao_ao_sparse(
            ao[0], aow1, None, nbins, None, pair_mask, ao_loc, hermi=0, out=vmat
        )
    else:
        vmat = _dot_ao_ao_sparse(
            ao[0], aow1, None, nbins, mask, pair_mask, ao_loc, hermi=0, out=vmat
        )

    if ndd > 0 or nd1 > 0:
        for j in range(3):
            i = j + 1
            # TODO might be able to use some kind of screening here
            aow1 = _scale_ao_sparse(ao[i], wv[4], mask, ao_loc, out=aow1)
            for ik in range(nk1):
                ir = 3 * ik + i + tau_idx + nk0
                indk = flsettings.get_ik0(ik)
                aow2 = _scale_ao_sparse(kao[indk], wv[ir], None, ao_loc, out=aow2)
                aow1 += aow2
            for ik in range(ndd):
                ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
                indk = flsettings.get_ik0(ik) + i
                aow2 = _scale_ao_sparse(kao[indk], wv[ir], None, ao_loc, out=aow2)
                aow1 += aow2
            vmat = _dot_ao_ao_sparse(
                ao[i], aow1, None, nbins, None, pair_mask, ao_loc, hermi=0, out=vmat
            )
    elif nk1 > 0:
        kwv = np.empty((nk1, wv.shape[-1]))
        gkao = kao[:nk1]
        for j in range(3):
            i = j + 1
            # TODO might be able to use some kind of screening here
            kwv[:] = wv[5 + nk0 + j : 5 + nk0 + 3 * nk1 : 3]
            aow1 = _scale_ao_sparse(ao[i], wv[4], mask, ao_loc, out=aow1)
            # aow2 = _scale_ao_sparse(gkao, kwv, mask, ao_loc, out=aow2)
            # aow1 += aow2
            # vmat = _dot_ao_ao_sparse(ao[i], aow1, None, nbins, mask, pair_mask,
            #                         ao_loc, hermi=0, out=vmat)
            aow2 = _scale_ao_sparse(gkao, kwv, None, ao_loc, out=aow2)
            aow1 += aow2
            vmat = _dot_ao_ao_sparse(
                ao[i], aow1, None, nbins, None, pair_mask, ao_loc, hermi=0, out=vmat
            )
    else:
        for j in range(3):
            i = j + 1
            aow1 = _scale_ao_sparse(ao[i], wv[4], mask, ao_loc, out=aow1)
            v1 = _dot_ao_ao_sparse(
                ao[i], aow1, None, nbins, mask, pair_mask, ao_loc, hermi=1, out=v1
            )

    return aow1, aow2, vmat, v1


def eval_rho2(
    mol,
    all_ao,
    mo_coeff,
    mo_occ,
    flsettings,
    non0tab=None,
    xctype="LDA",
    with_lapl=True,
    verbose=None,
):
    ao, kao = all_ao
    xctype = xctype.upper()
    ngrids, nao = ao.shape[-2:]

    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()
    pos = mo_occ > OCCDROP
    if np.any(pos):
        cpos = np.einsum("ij,j->ij", mo_coeff[:, pos], np.sqrt(mo_occ[pos]))
        if with_lapl:
            raise NotImplementedError
        else:
            rho = np.empty((5 + flsettings.nrho, ngrids))
            tau_idx = 4
        c0 = _dot_ao_dm(mol, ao[0], cpos, non0tab, shls_slice, ao_loc)
        # rho[0] = numpy.einsum('pi,pi->p', c0, c0)
        rho[0] = _contract_rho(c0, c0)
        cklst = []
        for i in range(kao.shape[0]):
            # TODO might be able to use some kind of screening here
            cklst.append(
                _dot_ao_dm(mol, kao[i], cpos, None, shls_slice, ao_loc)
                # _dot_ao_dm(mol, kao[i], cpos, non0tab, shls_slice, ao_loc)
            )
        nk0 = flsettings.nk0
        nk1 = flsettings.nk1
        nd1 = flsettings.nd1
        for ik in range(nk0):
            ir = ik + 1 + tau_idx
            indk = flsettings.get_ik0(ik)
            rho[ir] = _contract_rho(cklst[indk], c0)

        for ik in range(nd1):
            ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * ik)
            indk = flsettings.get_ik0(ik)
            for i in range(3):
                rho[ir + i] = _contract_rho(c0, cklst[indk + 1 + i])

        rho[tau_idx] = 0
        for ik in range(flsettings.ndd):
            ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
            rho[ir] = 0
        for i in range(1, 4):
            c1 = _dot_ao_dm(mol, ao[i], cpos, non0tab, shls_slice, ao_loc)
            rho[i] = _contract_rho(c0, c1) * 2
            rho[tau_idx] += _contract_rho(c1, c1)

            for ik in range(flsettings.nk1):
                ir = 3 * ik + i + tau_idx + nk0
                indk = flsettings.get_ik0(ik)
                rho[ir] = _contract_rho(cklst[indk], c1)

            for ik in range(flsettings.ndd):
                ir = (3 * nk1 + nk0) + (1 + tau_idx) + (3 * nd1) + ik
                indk = flsettings.get_ik0(ik) + i
                rho[ir] += _contract_rho(cklst[indk], c1)

        rho[tau_idx] *= 0.5
        return rho
    else:
        rho = np.zeros((5 + flsettings.nrho, ngrids))
        return rho


def _block_loop(
    ni,
    mol,
    grids,
    nao=None,
    deriv=0,
    max_memory=2000,
    non0tab=None,
    blksize=None,
    buf=None,
):
    """Define this macro to loop over grids by blocks."""
    if grids.coords is None:
        grids.build(with_non0tab=True)
    if nao is None:
        nao = mol.nao
    ngrids = grids.coords.shape[0]
    comp = (deriv + 1) * (deriv + 2) * (deriv + 3) // 6
    flsettings = ni.plan.settings
    kcomp = flsettings.npow + 3 * flsettings.nd1
    # NOTE to index grids.non0tab, the blksize needs to be an integer
    # multiplier of BLKSIZE
    if blksize is None:
        blksize = int(max_memory * 1e6 / ((comp + kcomp + 1) * nao * 8 * BLKSIZE))
        blksize = max(4, min(blksize, ngrids // BLKSIZE + 1, 1200)) * BLKSIZE
    assert blksize % BLKSIZE == 0

    if non0tab is None and mol is grids.mol:
        non0tab = grids.non0tab
    if non0tab is None:
        non0tab = np.empty(
            ((ngrids + BLKSIZE - 1) // BLKSIZE, mol.nbas), dtype=np.uint8
        )
        non0tab[:] = NBINS + 1  # Corresponding to AO value ~= 1
    screen_index = non0tab

    # the xxx_sparse() functions require ngrids 8-byte aligned
    allow_sparse = ngrids % ALIGNMENT_UNIT == 0

    if buf is None:
        buf = _empty_aligned((comp + kcomp) * blksize * nao)
    for ip0, ip1 in lib.prange(0, ngrids, blksize):
        coords = grids.coords[ip0:ip1]
        weight = grids.weights[ip0:ip1]
        mask = screen_index[ip0 // BLKSIZE :]
        # TODO: pass grids.cutoff to eval_ao
        ao_buf = np.ndarray((comp + kcomp, weight.size * nao), buffer=buf)
        ao = ni.eval_ao(
            mol, coords, deriv=deriv, non0tab=mask, cutoff=grids.cutoff, out=ao_buf
        )
        kao = ni.eval_kao(
            flsettings.slist,
            mol,
            coords,
            deriv=0,
            non0tab=None,
            cutoff=grids.cutoff,
            out=ao_buf[comp:],
            n1=flsettings.nd1,
        )
        if not allow_sparse and not _sparse_enough(mask):
            # Unset mask for dense AO tensor. It determines which eval_rho
            # to be called in make_rho
            mask = None
        yield (ao, kao), mask, weight, coords


def eval_kao(
    slst,
    mol,
    coords,
    deriv=0,
    shls_slice=None,
    non0tab=None,
    cutoff=None,
    out=None,
    n1=0,
    verbose=None,
):
    assert deriv == 0
    comp = len(slst)
    if mol.cart:
        "GTOval_cart_deriv%d" % deriv
    else:
        "GTOval_sph_deriv%d" % deriv
    if n1 == 0:
        return eval_flapl_gto(
            slst, mol, coords, shls_slice, non0tab, cutoff=cutoff, out=out
        )
    else:
        comp += 3 * n1
        nao = mol.nao_nr()
        ngrids = coords.shape[0]
        if out is None:
            out = np.ndarray((comp, nao, ngrids), buffer=out)
        else:
            assert out.shape[0] == comp
            assert out.flags.c_contiguous
            out = out.reshape((comp, nao, ngrids))
        eval_flapl_gto(
            slst[:n1],
            mol,
            coords,
            shls_slice,
            non0tab,
            cutoff=cutoff,
            out=out[: 4 * n1],
            deriv=1,
        )
        eval_flapl_gto(
            slst[n1:],
            mol,
            coords,
            shls_slice,
            non0tab,
            cutoff=cutoff,
            out=out[4 * n1 :],
        )
        return out.transpose(0, 2, 1)


def nr_rks(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    if relativity != 0:
        raise NotImplementedError

    xctype = "MGGA"

    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()
    cutoff = grids.cutoff * 1e2
    nbins = NBINS * 2 - int(NBINS * np.log(cutoff) / np.log(grids.cutoff))

    nelec = np.zeros(nset)
    excsum = np.zeros(nset)
    vmat = np.zeros((nset, nao, nao))

    def block_loop(ao_deriv):
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory
        ):
            for i in range(nset):
                rho = make_rho(i, ao, mask, xctype)
                exc, vxc = ni.eval_xc_eff(xc_code, rho, deriv=1, xctype=xctype)[:2]
                den = rho[0] * weight
                nelec[i] += den.sum()
                excsum[i] += np.dot(den, exc)
                wv = weight * vxc
                yield i, ao, mask, wv

    aow1 = None
    aow2 = None
    pair_mask = mol.get_overlap_cond() < -np.log(ni.cutoff)
    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    ao_deriv = 1
    v1 = np.zeros_like(vmat)
    for i, ao, mask, wv in block_loop(ao_deriv):
        _odp_dot_sparse_(
            ao,
            wv,
            nbins,
            mask,
            pair_mask,
            ao_loc,
            ni.flsettings,
            aow1=aow1,
            aow2=aow2,
            vmat=vmat[i],
            v1=v1[i],
        )
    vmat = lib.hermi_sum(vmat, axes=(0, 2, 1))
    vmat += v1


class FLNumInt(_NumIntMixin):
    """Numerical integration methods for non-relativistic RKS and UKS
    with fractional Laplacian density descriptors."""

    cutoff = CUTOFF * 1e2  # cutoff for small AO product

    def __init__(self, plan):
        """
        Initialize FLNumInt object.

        Args:
            plan (FracLaplPlan): Plan object for how to evaluate fractional
                Laplacian features
        """
        super(FLNumInt, self).__init__()
        self.plan = plan

    def nr_vxc(
        self,
        mol,
        grids,
        xc_code,
        dms,
        spin=0,
        relativity=0,
        hermi=0,
        max_memory=2000,
        verbose=None,
    ):
        if spin == 0:
            return self.nr_rks(
                mol, grids, xc_code, dms, relativity, hermi, max_memory, verbose
            )
        else:
            return self.nr_uks(
                mol, grids, xc_code, dms, relativity, hermi, max_memory, verbose
            )

    get_vxc = nr_vxc

    nr_rks = nr_rks
    nr_uks = None  # TODO write nr_uks

    # nr_nlc_vxc = numint.nr_nlc_vxc
    nr_sap = nr_sap_vxc = numint.nr_sap_vxc

    make_mask = staticmethod(make_mask)
    eval_ao = staticmethod(numint.eval_ao)
    eval_kao = staticmethod(eval_kao)
    eval_rho = staticmethod(eval_rho)
    eval_rho1 = lib.module_method(eval_rho1, absences=["cutoff"])
    eval_rho2 = staticmethod(eval_rho2)
    get_rho = numint.get_rho

    block_loop = _block_loop

    def _gen_rho_evaluator(self, mol, dms, hermi=0, with_lapl=True, grids=None):
        if getattr(dms, "mo_coeff", None) is not None:
            # TODO: test whether dm.mo_coeff matching dm
            mo_coeff = dms.mo_coeff
            mo_occ = dms.mo_occ
            if isinstance(dms, np.ndarray) and dms.ndim == 2:
                mo_coeff = [mo_coeff]
                mo_occ = [mo_occ]
        else:
            mo_coeff = mo_occ = None

        if isinstance(dms, np.ndarray) and dms.ndim == 2:
            dms = dms[np.newaxis]

        if hermi != 1 and dms[0].dtype == np.double:
            # (D + D.T)/2 because eval_rho computes 2*(|\nabla i> D_ij <j|) instead of
            # |\nabla i> D_ij <j| + |i> D_ij <\nabla j| for efficiency when dm is real
            dms = lib.hermi_sum(np.asarray(dms, order="C"), axes=(0, 2, 1)) * 0.5
            hermi = 1

        nao = dms[0].shape[0]
        ndms = len(dms)

        ovlp_cond = mol.get_overlap_cond()
        if dms[0].dtype == np.double:
            dm_cond = np.max(
                [mol.condense_to_shell(dm, "absmax") for dm in dms], axis=0
            )
            pair_mask = (np.exp(-ovlp_cond) * dm_cond) > self.cutoff
        else:
            pair_mask = ovlp_cond < -np.log(self.cutoff)

        def make_rho(idm, ao, sindex, xctype):
            if sindex is not None and grids is not None:
                return self.eval_rho1(
                    mol,
                    ao,
                    dms[idm],
                    self.plan.settings,
                    sindex,
                    xctype,
                    hermi,
                    with_lapl,
                    cutoff=self.cutoff,
                    ao_cutoff=grids.cutoff,
                    pair_mask=pair_mask,
                )
            elif mo_coeff is not None:
                return self.eval_rho2(
                    mol,
                    ao,
                    mo_coeff[idm],
                    mo_occ[idm],
                    self.plan.settings,
                    sindex,
                    xctype,
                    with_lapl,
                )
            else:
                return self.eval_rho(
                    mol,
                    ao,
                    dms[idm],
                    self.plan.settings,
                    sindex,
                    xctype,
                    hermi,
                    with_lapl,
                )

        return make_rho, ndms, nao



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np
from gpaw.atom.radialgd import RadialGridDescriptor
from pyscf.dft.radi import BRAGG_RADII

from ciderpress.lib import load_library as load_cider_library

cutil = load_cider_library("libsbt")
cutil.spherical_bessel_transform_setup.restype = ctypes.c_void_p


def _interp_helper(f_g, dg):
    N = f_g.shape[-1]
    g = dg.astype(int)
    g = np.minimum(g, N - 2)
    g = np.maximum(g, 0)
    dg -= g
    return (1 - dg) * f_g[..., g] + dg * f_g[..., g + 1]


class SBTRadialGridDescriptor(RadialGridDescriptor):
    def __init__(self, a, d, N=1000, default_spline_points=25):
        """Radial grid descriptor for Spherical Bessel Transform
        using the NUMSBT algorithm.

        The radial grid is::

                      dg
            r(g) = a(e  ),  g = 0, 1, ..., N - 1
        """

        self.a = a
        self.d = d
        g = np.arange(N)
        r_g = a * np.exp(d * g)
        dr_g = r_g * d
        RadialGridDescriptor.__init__(self, r_g, dr_g, default_spline_points)

    def r2g(self, r):
        g = np.zeros_like(r)
        g[r <= self.a] = 0  # NOTE: truncates g
        g[r > self.a] = np.log(r[r > self.a] / self.a) / self.d
        return g

    def d2gdr2(self):
        raise NotImplementedError

    def interpolate(self, f_g, r_x):
        return _interp_helper(f_g, self.r2g(r_x))

    def interpolate_self(self, f_g, rgd, zero_extrap=False):
        if zero_extrap:
            N = rgd.r_g.shape[0]
            g = rgd.r2g(self.r_g)
            res = _interp_helper(f_g, g)
            res[g > N - 1] = 0
            return res
        else:
            return _interp_helper(f_g, rgd.r2g(self.r_g))

    def new(self, N):
        return SBTRadialGridDescriptor(self.a, self.d, N)


class SBTFullGridDescriptor(SBTRadialGridDescriptor):
    def __init__(self, a_g, encut, d, N=1000, lmax=4, default_spline_points=25):
        """Radial grid descriptor for Spherical Bessel Transform
        using the NUMSBT algorithm.

        The radial grid is::

                      dg
            r(g) = a(e  ),  g = 0, 1, ..., N - 1
        """
        self.a = a_g
        self.d = d
        self.lmax = lmax
        g = np.arange(N)
        r_g = a_g * np.exp(d * g)
        dr_g = r_g * d
        self.k_g = np.zeros_like(r_g)

        self.sbt_desc = cutil.spherical_bessel_transform_setup(
            ctypes.c_double(encut),
            ctypes.c_int(lmax),
            ctypes.c_int(N),
            r_g.ctypes.data_as(ctypes.c_void_p),
            self.k_g.ctypes.data_as(ctypes.c_void_p),
        )

        self.a_k = self.k_g[0]
        self.dk_g = self.k_g * d

        RadialGridDescriptor.__init__(self, r_g, dr_g, default_spline_points)

    def __del__(self):
        cutil.free_sbt_descriptor(ctypes.c_void_p(self.sbt_desc))

    def k2g(self, r):
        g = np.zeros_like(r)
        g[r <= self.a_k] = 0  # NOTE: truncates g
        g[r > self.a_k] = np.log(r[r > self.a_k] / self.a_k) / self.d
        return g

    def interpolate_ks(self, f_k, k_x):
        return _interp_helper(f_k, self.k2g(k_x))

    def transform_single_fwd(self, f_g, l, f_k=None, l_add=0):
        if f_k is None:
            f_k = np.empty_like(f_g)
        cutil.wave_spherical_bessel_transform(
            ctypes.c_void_p(self.sbt_desc),
            f_g.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(l),
            f_k.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(l_add),
        )
        return f_k

    def transform_single_bwd(self, f_k, l, f_g=None):
        if f_g is None:
            f_g = np.empty_like(f_k)
        cutil.inverse_wave_spherical_bessel_transform(
            ctypes.c_void_p(self.sbt_desc),
            f_k.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(l),
            f_g.ctypes.data_as(ctypes.c_void_p),
        )
        return f_g

    def transform_set_fwd(self, f_xLg, f_xLk=None, l_add=0):
        Lmax = f_xLg.shape[-2]
        assert Lmax <= (self.lmax + 1) * (self.lmax + 1)
        nk = f_xLg.shape[-1]
        nset = f_xLg.size // nk
        f_xLg = np.ascontiguousarray(f_xLg)
        if f_xLk is None:
            f_xLk = np.zeros_like(f_xLg, order="C")
        else:
            f_xLk = np.ascontiguousarray(f_xLk)
        cutil.transform_set_fwd(
            ctypes.c_void_p(self.sbt_desc),
            f_xLg.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(Lmax),
            ctypes.c_int(nset),
            ctypes.c_int(nk),
            f_xLk.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(l_add),
        )
        return f_xLk

    def transform_set_bwd(self, f_xLk, f_xLg=None):
        Lmax = f_xLk.shape[-2]
        assert Lmax <= (self.lmax + 1) * (self.lmax + 1)
        nk = f_xLk.shape[-1]
        nset = f_xLk.size // nk
        f_xLk = np.ascontiguousarray(f_xLk)
        if f_xLg is None:
            f_xLg = np.zeros_like(f_xLk, order="C")
        else:
            f_xLg = np.ascontiguousarray(f_xLg)
        cutil.transform_set_bwd(
            ctypes.c_void_p(self.sbt_desc),
            f_xLk.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(Lmax),
            ctypes.c_int(nset),
            ctypes.c_int(nk),
            f_xLg.ctypes.data_as(ctypes.c_void_p),
        )
        return f_xLg


# from pyscf
def becke(n, charge, *args, **kwargs):
    """Becke, JCP 88, 2547 (1988); DOI:10.1063/1.454033"""
    if charge == 1:
        rm = BRAGG_RADII[charge]
    else:
        rm = BRAGG_RADII[charge] * 0.5
    t, w = np.polynomial.chebyshev.chebgauss(n)
    r = (1 + t) / (1 - t) * rm
    w *= 2 / (1 - t) ** 2 * rm * np.sqrt(1 - t**2)
    return r, w, rm


class GCRadialGridDescriptor(RadialGridDescriptor):
    def __init__(self, Z, N=100, default_spline_points=25):
        r_g, dr_g, self.rm = becke(N, Z)
        r_g = np.flip(r_g)
        dr_g = np.flip(dr_g)
        self.npt = N
        RadialGridDescriptor.__init__(self, r_g, dr_g, default_spline_points)

    def r2g(self, r):
        t = (r - self.rm) / (r + self.rm)
        t = np.arccos(t) / np.pi * 2 * self.npt
        return (self.npt - 1) - (t - 1) / 2

    def derivative(self, f, o=None):
        raise NotImplementedError

    def derivative2(self, f, o=None):
        raise NotImplementedError

    def make_cut(self, N):
        return CutGCRadialGridDescriptor(self.r_g[:N], self.dr_g[:N], self.rm, self.npt)


class CutGCRadialGridDescriptor(GCRadialGridDescriptor):
    def __init__(self, r_g, dr_g, rm, npt, default_spline_points=25):
        self.rm = rm
        self.npt = npt
        RadialGridDescriptor.__init__(self, r_g, dr_g, default_spline_points)



import numpy as np


def get_sigma(grad_svg):
    nspin = grad_svg.shape[0]
    shape = grad_svg.shape[2:]
    if nspin not in [1, 2]:
        raise ValueError("Only 1 or 2 spins supported")
    sigma_xg = np.empty((2 * nspin - 1,) + shape)
    sigma_xg[::2] = np.einsum("sv...,sv...->s...", grad_svg, grad_svg)
    if nspin == 2:
        sigma_xg[1] = np.einsum("v...,v...->...", grad_svg[0], grad_svg[1])
    return sigma_xg


def get_dsigmadf(grad_svg, dgraddf_ovg, spins):
    norb = len(spins)
    assert dgraddf_ovg.shape[0] == norb
    shape = grad_svg.shape[2:]
    dsigmadf_og = np.empty((norb,) + shape)
    for o, s in enumerate(spins):
        dsigmadf_og[o] = 2 * np.einsum("vg,vg->g", grad_svg[s], dgraddf_ovg[o])
    return dsigmadf_og



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np
from pyscf import gto
from pyscf.dft.gen_grid import (
    LEBEDEV_NGRID,
    LEBEDEV_ORDER,
    NELEC_ERROR_TOL,
    Grids,
    _default_ang,
    _default_rad,
    _padding_size,
    arg_group_grids,
    libdft,
    logger,
    nwchem_prune,
    radi,
)

from ciderpress.dft.grids_indexer import AtomicGridsIndexer, libcider

CIDER_DEFAULT_LMAX = 10


class CiderGrids(Grids):
    """
    Attributes (in addition to those in pyscf Grids object):
        lmax (int): Maximum l value for integration of CIDER features
        nlm (int): (lmax + 1)**2
        grids_indexer (AtomicGridsIndexer): object to sort and index
            grids for CIDER features.
    """

    grids_indexer: AtomicGridsIndexer

    _keys = {"grids_indexer", "nlm", "lmax"}

    def __init__(self, mol, lmax=CIDER_DEFAULT_LMAX):
        super(CiderGrids, self).__init__(mol)
        # self.becke_scheme = becke_lko
        self.lmax = lmax
        self.nlm = (lmax + 1) * (lmax + 1)
        self.grids_indexer = None

    def gen_atomic_grids(
        self,
        mol,
        atom_grid=None,
        radi_method=None,
        level=None,
        prune=None,
        build_indexer=False,
        **kwargs
    ):
        """
        Same as gen_atomic_grids for PySCF grids, except it also
        uses the outputs of gen_atomic_grids_cider to initialize the
        grids_indexer object for CIDER feature routines.
        """
        if atom_grid is None:
            atom_grid = self.atom_grid
        if radi_method is None:
            radi_method = self.radi_method
        if level is None:
            level = self.level
        if prune is None:
            prune = self.prune
        (
            atom_grids_tab,
            lmax_tab,
            rad_loc_tab,
            ylm_tab,
            ylm_loc_tab,
            rad_tab,
            dr_tab,
        ) = gen_atomic_grids_cider(
            mol, atom_grid, self.radi_method, level, prune, **kwargs
        )
        if build_indexer:
            self.grids_indexer = AtomicGridsIndexer.from_tabs(
                mol, self.lmax, rad_loc_tab, ylm_loc_tab, rad_tab, ylm_tab
            )
        return atom_grids_tab

    def build(self, mol=None, with_non0tab=False, sort_grids=True, **kwargs):
        """
        Build the grids. Same as with PySCF grids, but also stores an idx_map
        in the grids_indexer object so that one can map between the atomic grids
        and sorted grids.
        """
        if mol is None:
            mol = self.mol
        if self.verbose >= logger.WARN:
            self.check_sanity()
        atom_grids_tab = self.gen_atomic_grids(
            mol,
            self.atom_grid,
            self.radi_method,
            self.level,
            self.prune,
            build_indexer=True,
            **kwargs
        )
        # TODO cleaner version of this way of calling VXCgen_grid_lko
        # tmp = libdft.VXCgen_grid
        # libdft.VXCgen_grid = libcider.VXCgen_grid_lko
        self.coords, self.weights = self.get_partition(
            mol, atom_grids_tab, self.radii_adjust, self.atomic_radii, self.becke_scheme
        )
        # libdft.VXCgen_grid = tmp
        self.grids_indexer.set_weights(self.weights)

        if sort_grids:
            idx = arg_group_grids(mol, self.coords)
            self.coords = self.coords[idx]
            self.weights = self.weights[idx]
            self.grids_indexer.set_idx(idx)
        else:
            self.grids_indexer.set_idx(np.arange(self.weights.size))

        if self.alignment > 1:
            padding = _padding_size(self.size, self.alignment)
            logger.debug(self, "Padding %d grids", padding)
            if padding > 0:
                self.coords = np.vstack(
                    [self.coords, np.repeat([[1e-4] * 3], padding, axis=0)]
                )
                self.weights = np.hstack([self.weights, np.zeros(padding)])
                self.grids_indexer.set_padding(padding)

        self.coords = np.asarray(self.coords, order="C")
        self.weights = np.asarray(self.weights, order="C")

        if with_non0tab:
            self.non0tab = self.make_mask(mol, self.coords)
            self.screen_index = self.non0tab
        else:
            self.screen_index = self.non0tab = None
        logger.info(self, "tot grids = %d", len(self.weights))
        return self

    def reset(self, mol=None):
        self.grids_indexer = None
        return super().reset(mol=mol)

    def prune_by_density_(self, rho, threshold=0):
        """
        Prune grids if the electron density on the grid is small.
        Only difference from PySCF version is that grids_indexer.set_idx
        is called to make sure the map between unsorted and sorted/screened
        grids is correct.
        """
        if threshold == 0:
            return self

        mol = self.mol
        n = np.dot(rho, self.weights)
        if abs(n - mol.nelectron) < NELEC_ERROR_TOL * n:
            rho *= self.weights
            idx = abs(rho) > threshold / self.weights.size
            logger.debug(
                self, "Drop grids %d", self.weights.size - np.count_nonzero(idx)
            )
            self.coords = np.asarray(self.coords[idx], order="C")
            self.weights = np.asarray(self.weights[idx], order="C")
            old_idx = self.grids_indexer.get_idx()
            old_idx_size = old_idx.size
            self.grids_indexer.set_idx(old_idx[idx[:old_idx_size]])
            if self.alignment > 1:
                padding = _padding_size(self.size, self.alignment)
                logger.debug(self, "prune_by_density_: %d padding grids", padding)
                if padding > 0:
                    self.coords = np.vstack(
                        [self.coords, np.repeat([[1e-4] * 3], padding, axis=0)]
                    )
                    self.weights = np.hstack([self.weights, np.zeros(padding)])
                self.grids_indexer.set_padding(padding)
            self.non0tab = self.make_mask(mol, self.coords)
            self.screen_index = self.non0tab
            self.coords = np.asarray(self.coords, order="C")
            self.weights = np.asarray(self.weights, order="C")
        return self


LMAX_DICT = {v: k // 2 for k, v in LEBEDEV_ORDER.items()}


def gen_atomic_grids_cider(
    mol,
    atom_grid=None,
    radi_method=radi.gauss_chebyshev,
    level=3,
    prune=nwchem_prune,
    full_lmax=CIDER_DEFAULT_LMAX,
    **kwargs
):
    if atom_grid is None:
        atom_grid = {}
    if isinstance(atom_grid, (list, tuple)):
        atom_grid = dict([(mol.atom_symbol(ia), atom_grid) for ia in range(mol.natm)])
    atom_grids_tab = {}
    lmax_tab = {}
    rad_loc_tab = {}
    ylm_tab = {}
    ylm_loc_tab = {}
    rad_tab = {}
    dr_tab = {}
    for ia in range(mol.natm):
        symb = mol.atom_symbol(ia)

        if symb not in atom_grids_tab:
            chg = gto.charge(symb)
            if symb in atom_grid:
                n_rad, n_ang = atom_grid[symb]
                if n_ang not in LEBEDEV_NGRID:
                    raise ValueError("Unsupported angular grids %d" % n_ang)
            else:
                n_rad = _default_rad(chg, level)
                n_ang = _default_ang(chg, level)
            rad, dr = radi_method(n_rad, chg, ia, **kwargs)

            rad_weight = 4 * np.pi * rad**2 * dr

            if callable(prune):
                angs = prune(chg, rad, n_ang)
            else:
                angs = [n_ang] * n_rad
            logger.debug(
                mol, "atom %s rad-grids = %d, ang-grids = %s", symb, n_rad, angs
            )

            angs = np.array(angs)
            coords = []
            vol = []
            rad_loc = np.array([0], dtype=np.int32)
            lmaxs = np.array([LMAX_DICT[ang] for ang in angs]).astype(np.int32)
            lmaxs = np.minimum(lmaxs, full_lmax)
            nlm = (full_lmax + 1) * (full_lmax + 1)
            ylm_full = np.empty((0, nlm), order="C")
            ylm_loc = []
            rads = []
            drs = []
            for n in sorted(set(angs)):
                grid = np.empty((n, 4))
                libdft.MakeAngularGrid(
                    grid.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(n)
                )
                idx = np.where(angs == n)[0]
                yloc_curr = ylm_full.shape[0]
                ylm = np.zeros((n, nlm), order="C")
                sphgd = np.ascontiguousarray(grid[:, :3])
                libcider.recursive_sph_harm_vec(
                    ctypes.c_int(nlm),
                    ctypes.c_int(n),
                    sphgd.ctypes.data_as(ctypes.c_void_p),
                    ylm.ctypes.data_as(ctypes.c_void_p),
                )
                lmax_shl = LMAX_DICT[n]
                nlm_shl = (lmax_shl + 1) * (lmax_shl + 1)
                ylm[:, nlm_shl:] = 0.0
                ylm_full = np.append(ylm_full, ylm, axis=0)
                coords.append(
                    np.einsum("i,jk->ijk", rad[idx], grid[:, :3]).reshape(-1, 3)
                )
                vol.append(np.einsum("i,j->ij", rad_weight[idx], grid[:, 3]).ravel())
                rads.append(rad[idx])
                drs.append(dr[idx])
                rad_loc = np.append(
                    rad_loc, rad_loc[-1] + n * np.arange(1, len(idx) + 1)
                )
                ylm_loc.append(yloc_curr * np.ones(idx.size, dtype=np.int32))
            atom_grids_tab[symb] = (np.vstack(coords), np.hstack(vol))
            lmax_tab[symb] = lmaxs
            rad_loc_tab[symb] = rad_loc
            ylm_tab[symb] = ylm_full
            ylm_loc_tab[symb] = np.concatenate(ylm_loc).astype(np.int32)
            rad_tab[symb] = np.concatenate(rads).astype(np.float64)
            dr_tab[symb] = np.concatenate(drs).astype(np.float64)

    return atom_grids_tab, lmax_tab, rad_loc_tab, ylm_tab, ylm_loc_tab, rad_tab, dr_tab



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np

try:
    from gpaw import cgpaw
except ImportError:
    import _gpaw as cgpaw
from gpaw.grid_descriptor import GridDescriptor
from gpaw.xc.libvdwxc import FFTDistribution, nulltimer

from ciderpress.lib import c_null_ptr, load_library

pwutil = load_library("libpwutil")
pwutil.ciderpw_get_work_pointer.restype = ctypes.POINTER(ctypes.c_double)
pwutil.ciderpw_has_mpi.restype = ctypes.c_int


def get_norms_and_expnts_from_plan(plan):
    nalpha = plan.nalpha
    alphas = plan.alphas
    alpha_norms = plan.alpha_norms
    settings = plan.nldf_settings
    version = settings.version
    if version in ["j", "ij", "k"]:
        nbeta = nalpha
    else:
        nbeta = 0
    norms_ab = np.empty((nalpha, nbeta))
    expnts_ab = np.empty((nalpha, nbeta))
    t = 0
    if version in ["j", "ij", "k"]:
        aexp = alphas + alphas[:, None]
        fac = (np.pi / aexp) ** 1.5
        norms_ab[:, :nalpha] = fac * alpha_norms * alpha_norms[:, None]
        expnts_ab[:, :nalpha] = 1.0 / (4 * aexp)
        t = nalpha
    if version in ["i", "ij"]:
        for spec in settings.l0_feat_specs:
            fac = alpha_norms * (np.pi / alphas) ** 1.5
            # TODO this won't be correct for r^2 and lapl ones.
            if spec == "se":
                fac *= 1
            elif spec == "se_r2":
                fac *= 1
            elif spec == "se_apr2":
                fac *= alphas
            elif spec == "se_ap":
                fac *= alphas
            elif spec == "se_ap2r2":
                fac *= alphas * alphas
            elif spec == "se_lapl":
                fac *= 1
            else:
                raise ValueError
            norms_ab[:, t] = fac
            expnts_ab[:, t] = 1.0 / (4 * alphas)
            t += 1
        for spec in settings.l1_feat_specs:
            fac = alpha_norms * (np.pi / alphas) ** 1.5
            # TODO might be rt(3) or 1/2 prefactor or something
            if spec == "se_grad":
                fac *= 1.0
            elif spec == "se_rvec":
                fac /= alphas
            else:
                raise ValueError
            norms_ab[:, t : t + 3] = fac[:, None]
            expnts_ab[:, t : t + 3] = alphas[:, None]
            t += 3
    return nalpha, nbeta, alphas, alpha_norms, norms_ab, expnts_ab


class LibCiderPW:
    """
    This class provides an interface to the C-level, FFT-based evaluator
    routines for nonlocal density features (NLDFs).
    """

    def __init__(self, N_c, cell_cv, comm, plan=None):
        """
        Args:
            N_c (tuple): number of grid points along each lattice vector
            cell_cv (np.ndarray): 3x3 lattice vector matrix. Each row
                should be a lattice vector.
            comm: MPI communicator object
            plan (NLDFPlan): A plan object containing information about the
                convolutions to be performed to evaluate the NLDFs.
        """
        self.initialized = False
        ptr = c_null_ptr()
        self.shape = tuple(N_c)
        if plan is None:
            self.alpha_norms = np.ones(1)
            self.alphas = np.ones(1)
            self.nalpha = 1
            self.nbeta = 1
            self.norms_ab = np.ones((1, 1))
            self.expnts_ab = np.ones((1, 1))
        else:
            (
                self.nalpha,
                self.nbeta,
                self.alphas,
                self.alpha_norms,
                self.norms_ab,
                self.expnts_ab,
            ) = get_norms_and_expnts_from_plan(plan)
        alphas = np.asarray(self.alphas, dtype=np.float64, order="C")
        alpha_norms = np.asarray(self.alpha_norms, dtype=np.float64, order="C")
        assert len(alphas) == len(alpha_norms)
        assert cell_cv.flags.c_contiguous
        pwutil.ciderpw_create(
            ctypes.byref(ptr),
            ctypes.c_double(1),
            np.array(N_c).astype(np.int32).ctypes.data_as(ctypes.c_void_p),
            cell_cv.astype(np.float64).ctypes.data_as(ctypes.c_void_p),
        )
        self._ptr = ptr
        self.comm = comm
        self._local_shape_real = None
        self._local_shape_recip = None
        self._glda = None
        self._klda = None
        self._data_arr = None
        if not self.has_mpi and self.comm.size > 1:
            raise ValueError(
                "CIDER is only compiled for serial evaluation, but the "
                "calculation is parallel with {} processes.".format(self.comm.size)
            )

    @property
    def has_mpi(self):
        """
        Check whether libpwutil.so was compiled with MPI support.
        """
        return bool(int(pwutil.ciderpw_has_mpi()))

    def __del__(self):
        """
        We need to free the C memory was the Python object is deleted.
        """
        pwutil.ciderpw_finalize(ctypes.byref(self._ptr))
        self._ptr = None

    def initialize_backend(self):
        """
        Set up all the data and memory allocations for the C backend.
        """
        assert not self.initialized

        if self.has_mpi:
            comm = self.comm
            pwutil.ciderpw_init_mpi_from_gpaw(
                self._ptr,
                ctypes.py_object(comm.get_c_object()),
                ctypes.c_int(self.nalpha),
                ctypes.c_int(self.nbeta),
                self.norms_ab.ctypes.data_as(ctypes.c_void_p),
                self.expnts_ab.ctypes.data_as(ctypes.c_void_p),
            )
        else:
            pwutil.ciderpw_init_serial(
                self._ptr,
                ctypes.c_int(self.nalpha),
                ctypes.c_int(self.nbeta),
                self.norms_ab.ctypes.data_as(ctypes.c_void_p),
                self.expnts_ab.ctypes.data_as(ctypes.c_void_p),
            )

        size_data = np.zeros(8, dtype=np.int32)
        pwutil.ciderpw_get_local_size_and_lda(
            self._ptr, size_data.ctypes.data_as(ctypes.c_void_p)
        )
        self._local_shape_real = size_data[0:3]
        self._glda = size_data[3]
        self._local_shape_recip = size_data[4:7]
        self._klda = size_data[7]
        self.initialized = True

    def fftv2(self, arr_gq):
        pwutil.ciderpw_g2k_v2(
            self._ptr,
            arr_gq.ctypes.data_as(ctypes.c_void_p),
        )

    def fft(self, in_xg, out_xg):
        if self.has_mpi:
            fn = pwutil.ciderpw_g2k_mpi_gpaw
        else:
            fn = pwutil.ciderpw_g2k_serial_gpaw
        for x in range(in_xg.shape[0]):
            fn(
                self._ptr,
                in_xg[x].ctypes.data_as(ctypes.c_void_p),
                out_xg[x].ctypes.data_as(ctypes.c_void_p),
            )

    def ifft(self, in_xg, out_xg):
        if self.has_mpi:
            fn = pwutil.ciderpw_k2g_mpi_gpaw
        else:
            fn = pwutil.ciderpw_k2g_serial_gpaw
        for x in range(in_xg.shape[0]):
            fn(
                self._ptr,
                in_xg[x].ctypes.data_as(ctypes.c_void_p),
                out_xg[x].ctypes.data_as(ctypes.c_void_p),
            )

    def compute_forward_convolution(self):
        pwutil.ciderpw_compute_features(self._ptr)

    def compute_backward_convolution(self):
        pwutil.ciderpw_compute_potential(self._ptr)

    def compute_backward_convolution_with_stress(self, theta_gq):
        stress_vv = np.zeros((3, 3))
        pwutil.ciderpw_convolution_potential_and_stress(
            self._ptr,
            stress_vv.ctypes.data_as(ctypes.c_void_p),
            theta_gq.ctypes.data_as(ctypes.c_void_p),
        )
        return stress_vv

    def fill_vj_feature_(self, feat_g, p_gq):
        """
        This function is an IN-PLACE operation on feat_g. It takes the convolution
        weights p_gq for the feature convolutions, multiplies them by the
        convolution arrays work_gq stored in the FFTW C buffer, and contracts
        the q index, and sets feat_g to the result. feat_g values
        can be uninitialized on entry.

        Args:
            feat_g (np.ndarray): The array to write to. Must be C contiguous.
                On exit, feat_g will contain the sum over q of
                work_gq * p_gq, where work_gq is the internal FFTW C buffer.
            p_gq (np.ndarray): The convolution weights for this NLDF.
        """
        assert feat_g.flags.c_contiguous
        assert p_gq.flags.c_contiguous
        assert feat_g.shape == p_gq.shape[:-1] == tuple(self._local_shape_real)
        assert p_gq.shape[-1] == self.nalpha
        pwutil.ciderpw_eval_feature_vj(
            self._ptr,
            feat_g.ctypes.data_as(ctypes.c_void_p),
            p_gq.ctypes.data_as(ctypes.c_void_p),
        )

    def fill_vj_potential_(self, vfeat_g, p_gq):
        """
        This function is an IN-PLACE operation on the internal C work array.
        The basic effect of this function is work_gq += vfeat_g * p_gq.
        Note the +=. The internal work array should be zeroed using
        the reset_work() function before calling this with the first
        potential contribution.

        Args:
            vfeat_g (np.ndarray): Functional derivative of the energy with
                respect to an NLDF.
            p_gq (np.ndarray): The convolution weights for this feature.
        """
        assert vfeat_g.flags.c_contiguous
        assert p_gq.flags.c_contiguous
        assert vfeat_g.shape == p_gq.shape[:-1] == tuple(self._local_shape_real)
        assert p_gq.shape[-1] == self.nalpha
        pwutil.ciderpw_add_potential_vj(
            self._ptr,
            vfeat_g.ctypes.data_as(ctypes.c_void_p),
            p_gq.ctypes.data_as(ctypes.c_void_p),
        )

    def set_work(self, fun_g, p_gq):
        """
        Set the internal work array to work_gq = fun_g * p_gq.

        Args:
            fun_g (np.ndarray)
            p_gq (np.ndarray)
        """
        assert fun_g.flags.c_contiguous
        assert p_gq.flags.c_contiguous
        assert fun_g.shape == p_gq.shape[:-1] == tuple(self._local_shape_real)
        assert p_gq.shape[-1] == self.nalpha
        pwutil.ciderpw_set_work(
            self._ptr,
            fun_g.ctypes.data_as(ctypes.c_void_p),
            p_gq.ctypes.data_as(ctypes.c_void_p),
        )

    def reset_work(self):
        """
        Set the internal work array to zero.
        """
        pwutil.ciderpw_zero_work(self._ptr)

    def get_radial_info_for_atom(self, local_indices, local_fdisps):
        """
        Given 3D grid indices (local to the process) and the fractional
        coordinate displacements of these grids from an atom of interest,
        evaluate the radial distances and directions of each grid
        from the atom.

        Args:
            local_indices (np.ndarray): 3 x N array. Note, these
                indices must be local to the process. So if my process
                has grids 40-60 along lattice vector 1 and want the
                distance for grid 45, the 1st index should be 5.
            local_fdisps (np.ndarray): 3 x N array, fractional coordinates
                of the coordinates that correspond to the indices above.

        Returns:
            np.ndarray: The flattened, row-major indices corresponding
                to local_indices
            np.ndarray: The distances between each grid point and the atom.
            np.ndarray: The direction between each grid point and the atom.
                Has shape (N, 3).
        """
        num_c = []
        for i in range(3):
            num_c.append(len(local_indices[i]))
            assert len(local_fdisps[i]) == num_c[i]
        ngrid = np.prod(num_c)
        r_g = np.empty((4, ngrid), order="C")
        inds = np.ascontiguousarray(np.concatenate(local_indices).astype(np.int64))
        disps = np.ascontiguousarray(np.concatenate(local_fdisps).astype(np.float64))
        locs_g = np.empty(ngrid, dtype=np.int64)
        num_c = np.asarray(num_c, order="C", dtype=np.int64)
        pwutil.ciderpw_fill_atom_info(
            self._ptr,
            inds.ctypes.data_as(ctypes.c_void_p),
            disps.ctypes.data_as(ctypes.c_void_p),
            num_c.ctypes.data_as(ctypes.c_void_p),
            r_g.ctypes.data_as(ctypes.c_void_p),
            locs_g.ctypes.data_as(ctypes.c_void_p),
        )
        return locs_g, r_g[0], r_g[1:].T

    def add_paw2grid(self, funcs_gb, indset):
        """
        Add the data in funcs_gb to the work array at the indices
        given by indset. This is an IN-PLACE operation on the
        internal work array in the C backend.
        """
        assert funcs_gb.flags.c_contiguous
        assert funcs_gb.dtype == np.float64
        assert indset.flags.c_contiguous
        assert indset.dtype == np.int64
        pwutil.ciderpw_add_atom_info(
            self._ptr,
            funcs_gb.ctypes.data_as(ctypes.c_void_p),
            indset.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(indset.size),
        )

    def set_grid2paw(self, funcs_gb, indset):
        """
        Set the array funcs_gb (overwritten on exit) to the values
        of the internal work array at the indices given by indset.
        """
        assert funcs_gb.flags.c_contiguous
        assert indset.flags.c_contiguous
        pwutil.ciderpw_set_atom_info(
            self._ptr,
            funcs_gb.ctypes.data_as(ctypes.c_void_p),
            indset.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(indset.size),
        )

    def get_bound_inds(self):
        """
        Get the upper and lower boundary indices for the grid on this process.

        Returns:
            np.ndarray: Lower bounds (length 3)
            np.ndarray: Upper bounds (length 3)
        """
        bound_inds = np.zeros(6, dtype=np.int32)
        pwutil.ciderpw_get_bound_inds(
            self._ptr, bound_inds.ctypes.data_as(ctypes.c_void_p)
        )
        return bound_inds[:3], bound_inds[3:]

    def get_recip_size(self):
        """
        Returns the size of the reciprocal-space work array on this process.
        """
        pwutil.ciderpw_get_recip_size.restype = ctypes.c_size_t
        return np.uintp(pwutil.ciderpw_get_recip_size(self._ptr))

    def get_real_size(self):
        """
        Returns the size of the real-space work array on this process.
        """
        pwutil.ciderpw_get_real_size.restype = ctypes.c_size_t
        return np.uintp(pwutil.ciderpw_get_real_size(self._ptr))

    def get_work_size(self):
        """
        Returns the true size of the work array on this process. Should be
        the max of get_real_size and get_recip_size, but this gets the actual
        size of the allocated work array.
        """
        pwutil.ciderpw_get_work_size.restype = ctypes.c_size_t
        return np.uintp(pwutil.ciderpw_get_work_size(self._ptr))

    def copy_work_array(self):
        """
        Make a copy of the work array and return it. Note that this is the raw
        work array; it might contain some uninitialized buffer.
        """
        arr = np.empty(self.get_work_size() * self.nalpha, dtype=np.complex128)
        pwutil.ciderpw_copy_work_array(self._ptr, arr.ctypes.data_as(ctypes.c_void_p))
        return arr


class FFTWrapper:
    def __init__(self, fft_obj, distribution, pd, timer=nulltimer):
        self.fft_obj = fft_obj
        self.dist1 = distribution
        gd = pd.gd
        if fft_obj.has_mpi:
            nrecip_c = [gd.N_c[1], gd.N_c[0], gd.N_c[2] // 2 + 1]
            parsize_c = [gd.parsize_c[1], gd.parsize_c[0], gd.parsize_c[2]]
        else:
            nrecip_c = [gd.N_c[0], gd.N_c[1], gd.N_c[2] // 2 + 1]
            parsize_c = [gd.parsize_c[0], gd.parsize_c[1], gd.parsize_c[2]]
        self.dummy_gd = GridDescriptor(
            nrecip_c,
            comm=gd.comm,
            parsize_c=parsize_c,
        )
        self.dist2 = FFTDistribution(self.dummy_gd, [gd.comm.size, 1, 1])
        self.pd = pd
        self.timer = timer

    def get_reciprocal_vectors(self, q=0, add_q=True):
        return self.pd.get_reciprocal_vectors(q=q, add_q=True)

    @property
    def G2_qG(self):
        return self.pd.G2_qG

    def integrate(self, *args, **kwargs):
        return self.pd.integrate(*args, **kwargs)

    def zeros(self, x=(), dtype=None, q=None, global_array=False):
        return self.pd.zeros(x=x, dtype=dtype, q=q, global_array=global_array)

    def empty(self, x=(), dtype=None, q=None, global_array=False):
        return self.pd.empty(x=x, dtype=dtype, q=q, global_array=global_array)

    def fft(self, in_xg, q=None, Q_G=None, local=False):
        if q is not None or Q_G is not None or local:
            raise NotImplementedError
        if in_xg.ndim == 3:
            in_xg = in_xg[None, ...]
        assert in_xg.ndim == 4
        xshape = (len(in_xg),)
        inblock_xg = self.dist1.block_zeros(xshape)
        outblock_xg = self.dist2.block_zeros(xshape).astype(np.complex128)
        self.dist1.gd2block(in_xg, inblock_xg)
        self.fft_obj.fft(inblock_xg, outblock_xg)
        contribs = []
        out_xg = self.dummy_gd.zeros(n=xshape, dtype=complex, global_array=False)
        self.dist2.block2gd_add(outblock_xg, out_xg)
        pd = self.pd
        for x in range(in_xg.shape[0]):
            out_g = self.dummy_gd.collect(out_xg[x])
            if self.pd.gd.comm.rank == 0:
                if self.fft_obj.has_mpi:
                    out_g = np.ascontiguousarray(out_g.transpose(1, 0, 2))
                to_scatter = out_g.ravel()[pd.Q_qG[0]]
            else:
                to_scatter = None
            contribs.append(pd.scatter(to_scatter))
        return np.stack(contribs)

    def ifft(self, in_xk, q=None, Q_G=None, local=False):
        if q is not None or Q_G is not None or local:
            raise NotImplementedError
        if in_xk.ndim == 1:
            in_xk = in_xk[None, ...]
        assert in_xk.ndim == 2
        contribs = []
        pd = self.pd
        for in_k in in_xk:
            in_k = pd.gather(in_k)
            if in_k is not None:
                pd.tmp_Q[:] = 0
                scale = 1.0 / self.pd.tmp_R.size
                cgpaw.pw_insert(in_k, pd.Q_qG[0], scale, pd.tmp_Q)
                t = pd.tmp_Q[:, :, 0]
                n, m = pd.gd.N_c[:2] // 2 - 1
                t[0, -m:] = t[0, m:0:-1].conj()
                t[n:0:-1, -m:] = t[-n:, m:0:-1].conj()
                t[-n:, -m:] = t[n:0:-1, m:0:-1].conj()
                t[-n:, 0] = t[n:0:-1, 0].conj()
            if self.fft_obj.has_mpi:
                tmp = pd.tmp_Q.transpose(1, 0, 2).copy()
            else:
                tmp = pd.tmp_Q.copy()
            contribs.append(self.dummy_gd.distribute(tmp))
        in_xg = np.stack(contribs)
        xshape = (len(in_xg),)
        inblock_xg = self.dist2.block_zeros(xshape).astype(np.complex128)
        outblock_xg = self.dist1.block_zeros(xshape)
        self.dist2.gd2block(in_xg, inblock_xg)
        self.fft_obj.ifft(inblock_xg, outblock_xg)
        out_xg = self.pd.gd.zeros(n=xshape)
        self.dist1.block2gd_add(outblock_xg, out_xg)
        return out_xg


def wrap_fft_mpi(pd):
    parsize_c = [pd.gd.comm.size, 1, 1]
    distribution = FFTDistribution(pd.gd, parsize_c)
    parsize_c = [1, pd.gd.comm.size, 1]
    fft_obj = LibCiderPW(pd.gd.N_c, pd.gd.cell_cv, comm=pd.gd.comm)
    fft_obj.initialize_backend()
    wrapper = FFTWrapper(fft_obj, distribution, pd)
    return wrapper


if __name__ == "__main__":
    from gpaw.mpi import MPIComm, world

    np.random.seed(98)

    comm = MPIComm(world)
    N_c = [100, 80, 110]
    cell_cv = np.diag(np.ones(3))
    ciderpw = LibCiderPW(N_c, cell_cv, comm)
    ciderpw.initialize_backend()

    from gpaw.pw.descriptor import PWDescriptor

    gd = GridDescriptor(N_c, cell_cv, comm=MPIComm(world))
    pd = PWDescriptor(180, gd, dtype=float)

    wrapper = wrap_fft_mpi(pd)

    input = np.random.normal(size=N_c)[None, :, :, :]
    input = gd.distribute(input)

    output_ref = pd.fft(input)
    sum = np.abs(output_ref).sum()
    sum = pd.gd.comm.sum_scalar(sum)
    import time

    t0 = time.monotonic()
    for i in range(20):
        output_test = wrapper.fft(input)
    t1 = time.monotonic()
    print("TIME", t1 - t0)
    input_test = wrapper.ifft(output_test)
    input_ref = pd.ifft(output_ref)

    from numpy.testing import assert_allclose

    assert_allclose(output_test[0], output_ref)
    assert_allclose(input_test[0], input_ref)
